import logging
from typing import Any, Dict, List, Optional, Union

from Newmode import Client
from oauthlib.oauth2 import TokenExpiredError

from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.oauth_api_connector import OAuth2APIConnector

logger = logging.getLogger(__name__)

V2_API_URL = "https://base.newmode.net/api/"
V2_API_AUTH_URL = "https://base.newmode.net/oauth/token/"
V2_API_CAMPAIGNS_URL = "https://base.newmode.net/"
V2_API_CAMPAIGNS_VERSION = "jsonapi"
V2_API_CAMPAIGNS_HEADERS = {
    "content-type": "application/vnd.api+json",
    "accept": "application/vnd.api+json",
    "authorization": "Bearer 1234567890",
}

PAGINATION_NEXT = "next"
RESPONSE_DATA_KEY = "data"
RESPONSE_LINKS_KEY = "links"


class NewmodeV1:
    def __init__(
        self,
        api_user: Optional[str] = None,
        api_password: Optional[str] = None,
        api_version: Optional[str] = None,
    ):
        """
        Args:
            api_user: str
                The Newmode api user. Not required if ``NEWMODE_API_USER`` env variable is
                passed.
            api_password: str
                The Newmode api password. Not required if ``NEWMODE_API_PASSWORD`` env variable is
                passed.
            api_version: str
                The Newmode api version. Defaults to "v1.0" or the value of ``NEWMODE_API_VERSION``
                env variable.
        Returns:
            Newmode class
        """
        logger.warning(
            "Newmode V1 API will be sunset in Feburary 28th, 2025. To use V2, set api_version=v2.1"
        )
        self.api_user: str = check_env.check("NEWMODE_API_USER", api_user)
        self.api_password: str = check_env.check("NEWMODE_API_PASSWORD", api_password)
        self.api_version: Optional[str] = api_version
        self.client: Client = Client(api_user, api_password, api_version)

    def convert_to_table(self, data: Union[List[Dict[str, Any]], Dict[str, Any]]) -> Table:
        # Internal method to create a Parsons table from a data element.
        table = None
        if isinstance(data, list):
            table = Table(data)
        else:
            table = Table([data])

        return table

    def get_tools(self, params: Dict[str, Any] = None) -> Table:
        """
        Get existing tools.
        Args:
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Tools information as table.
        """
        if params is None:
            params = {}
        tools = self.client.getTools(params=params)
        if tools:
            return self.convert_to_table(tools)
        else:
            logging.warning("Empty tools returned")
            return self.convert_to_table([])

    def get_tool(
        self, tool_id: Union[int, str], params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """
        Get specific tool.
        Args:
            tool_id:
                The id of the tool to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Tool information.
        """
        if params is None:
            params = {}
        tool = self.client.getTool(tool_id, params=params)
        if tool:
            return tool
        else:
            logging.warning("Empty tool returned")
            return None

    def lookup_targets(
        self,
        tool_id: Union[int, str],
        search: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Table:
        """
        Lookup targets for a given tool
        Args:
            tool_id:
                The tool to lookup targets.
            search:
                The search criteria. It could be:
                - Empty: If empty, return custom targets associated to the tool.
                - Postal code: Return targets matched by postal code.
                - Lat/Long: Latitude and Longitude pair separated by '::'.
                Ex. 45.451596::-73.59912099999997. It will return targets
                matched for those coordinates.
                - Search term: For your csv tools, this will return targets
                matched by given valid search term.
        Returns:
            Targets information as table.
        """
        if params is None:
            params = {}
        targets = self.client.lookupTargets(tool_id, search, params=params)
        if targets:
            data = []
            for key in targets:
                if key != "_links":
                    data.append(targets[key])
            return self.convert_to_table(data)
        else:
            logging.warning("Empty targets returned")
            return self.convert_to_table([])

    def get_action(
        self, tool_id: Union[int, str], params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """
        Get the action information for a given tool.
        Args:
            tool_id:
                The id of the tool to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Tool action information.
        """
        if params is None:
            params = {}
        action = self.client.getAction(tool_id, params=params)
        if action:
            return action
        else:
            logging.warning("Empty action returned")
            return None

    def run_action(
        self,
        tool_id: Union[int, str],
        payload: Dict[str, Any],
        params: Optional[Dict[str, Any]] = None,
    ) -> Optional[Union[str, int]]:
        """
        Run specific action with given payload.
        Args:
            tool_id:
                The id of the tool to run.
            payload:
                Payload data used to run the action. Structure will depend
                on the stuff returned by get_action.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Action link (if otl) or sid.
        """
        if params is None:
            params = {}
        action = self.client.runAction(tool_id, payload, params=params)
        if action:
            if "link" in action:
                return action["link"]
            else:
                return action["sid"]
        else:
            logging.warning("Error in response")
            return None

    def get_target(
        self, target_id: Union[int, str], params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """
        Get specific target.
        Args:
            target_id:
                The id of the target to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Target information.
        """
        if params is None:
            params = {}
        target = self.client.getTarget(target_id, params=params)
        if target:
            return target
        else:
            logging.warning("Empty target returned")
            return None

    def get_targets(self, params: Optional[Dict[str, Any]] = None) -> Optional[Table]:
        """
        Get all targets

        Args:
            params dict:
                Extra paramaters sent to New/Mode library

        Returns:
            Target information
        """

        if params is None:
            params = {}
        targets = self.client.getTargets(params=params)

        if targets:
            return self.convert_to_table(targets)

        else:
            logging.warning("No targets returned")
            return None

    def get_campaigns(self, params: Optional[Dict[str, Any]] = None) -> Table:
        """
        Get existing campaigns.
        Args:
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Campaigns information as table.
        """
        if params is None:
            params = {}
        campaigns = self.client.getCampaigns(params=params)
        if campaigns:
            return self.convert_to_table(campaigns)
        else:
            logging.warning("Empty campaigns returned")
            return self.convert_to_table([])

    def get_campaign(
        self, campaign_id: Union[int, str], params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """
        Get specific campaign.
        Args:
            campaign_id:
                The id of the campaign to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Campaign information.
        """
        if params is None:
            params = {}
        campaign = self.client.getCampaign(campaign_id, params=params)
        if campaign:
            return campaign
        else:
            logging.warning("Empty campaign returned")
            return None

    def get_organizations(self, params: Optional[Dict[str, Any]] = None) -> Table:
        """
        Get existing organizations.
        Args:
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Organizations information as table.
        """
        if params is None:
            params = {}
        organizations = self.client.getOrganizations(params=params)
        if organizations:
            return self.convert_to_table(organizations)
        else:
            logging.warning("Empty organizations returned")
            return self.convert_to_table([])

    def get_organization(
        self, organization_id: Union[int, str], params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """
        Get specific organization.
        Args:
            organization_id:
                The id of the organization to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Organization information.
        """
        if params is None:
            params = {}
        organization = self.client.getOrganization(organization_id, params=params)
        if organization:
            return organization
        else:
            logging.warning("Empty organization returned")
            return None

    def get_services(self, params: Optional[Dict[str, Any]] = None) -> Table:
        """
        Get existing services.
        Args:
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Services information as table.
        """
        if params is None:
            params = {}
        services = self.client.getServices(params=params)
        if services:
            return self.convert_to_table(services)
        else:
            logging.warning("Empty services returned")
            return self.convert_to_table([])

    def get_service(
        self, service_id: Union[int, str], params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """
        Get specific service.
        Args:
            service_id:
                The id of the service to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Service information.
        """
        if params is None:
            params = {}
        service = self.client.getService(service_id, params=params)
        if service:
            return service
        else:
            logging.warning("Empty service returned")
            return None

    def get_outreaches(
        self, tool_id: Union[int, str], params: Optional[Dict[str, Any]] = None
    ) -> Table:
        """
        Get existing outreaches for a given tool.
        Args:
            tool_id:
                Tool to return outreaches.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Outreaches information as table.
        """
        if params is None:
            params = {}
        outreaches = self.client.getOutreaches(tool_id, params=params)
        if outreaches:
            return self.convert_to_table(outreaches)
        else:
            logging.warning("Empty outreaches returned")
            return self.convert_to_table([])

    def get_outreach(
        self, outreach_id: Union[int, str], params: Optional[Dict[str, Any]] = None
    ) -> Optional[Dict[str, Any]]:
        """
        Get specific outreach.
        Args:
            outreach_id:
                The id of the outreach to return.
            params:
                Extra parameters sent to New/Mode library.
        Returns:
            Outreach information.
        """
        if params is None:
            params = {}
        outreach = self.client.getOutreach(outreach_id, params=params)
        if outreach:
            return outreach
        else:
            logging.warning("Empty outreach returned")
            return None


class NewmodeV2:
    # TODO: Add param definition and requirements once official Newmode docs are published
    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        api_version: str = "v2.1",
    ):
        """
        Instantiate Class
        `Args`:
            client_id: str
                The client id to use for the API requests. Not required if ``NEWMODE_API_CLIENT_ID``
                env variable set.
            client_secret: str
                The client secret to use for the API requests. Not required if ``NEWMODE_API_CLIENT_SECRET``
                env variable set.
            api_version: str
                The api version to use. Defaults to v2.1
        Returns:
            NewMode Class
        """
        self.api_version: str = api_version
        self.base_url: str = V2_API_URL
        self.client_id: str = check_env.check("NEWMODE_API_CLIENT_ID", client_id)
        self.client_secret: str = check_env.check("NEWMODE_API_CLIENT_SECRET", client_secret)
        self.headers: Dict[str, str] = {"content-type": "application/json"}
        self.default_client: OAuth2APIConnector = self.get_default_oauth_client()

    def get_default_oauth_client(self) -> OAuth2APIConnector:
        return OAuth2APIConnector(
            uri=self.base_url,
            auto_refresh_url=V2_API_AUTH_URL,
            client_id=self.client_id,
            client_secret=self.client_secret,
            headers=self.headers,
            token_url=V2_API_AUTH_URL,
            grant_type="client_credentials",
        )

    def checked_response(
        self, response: Any, client: OAuth2APIConnector
    ) -> Optional[Dict[str, Any]]:
        response.raise_for_status()
        success_codes = [200, 201, 202, 204]
        client.validate_response(response)
        if response.status_code in success_codes:
            try:
                if client.json_check(response):
                    return response.json()
            except Exception:
                logger.error("Response is not in JSON format.")
        raise ValueError(f"API request encountered an error. Response: {response}")

    def base_request(
        self,
        method: str,
        url: str,
        client: OAuth2APIConnector,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        retries: int = 2,
    ) -> Optional[Union[Dict[str, Any], None]]:
        """
        Internal method to instantiate OAuth2APIConnector class,
        make a single call to Newmode API, and validate the response.
        """
        if params is None:
            params = {}

        for attempt in range(retries + 1):
            try:
                try:
                    response = client.request(
                        url=url, req_type=method, json=json, data=data, params=params
                    )
                    return self.checked_response(response, client)
                except TokenExpiredError as e:
                    logger.warning(f"Token expired: {e}. Refreshing it...")
                    self.default_client = self.get_default_oauth_client()
            except Exception as e:
                if attempt < retries:
                    logger.warning(f"Request failed (attempt {attempt + 1}/{retries}). Retrying...")
                else:
                    logger.error(f"Request failed after {retries} retries.")
                    raise e
        raise Exception(f"Failed to retrieve data from {url} after {retries} attempts.")

    def paginate_request(
        self,
        method: str,
        endpoint: str,
        client: OAuth2APIConnector,
        data_key: str = RESPONSE_DATA_KEY,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        supports_version: bool = True,
        override_api_version: Optional[str] = None,
        retries: int = 2,
    ) -> List[Dict[str, Any]]:
        """
        Wrapper method to handle pagination for API requests.
        """
        if params is None:
            params = {}
        results = []
        api_version = override_api_version if override_api_version else self.api_version
        url = f"{api_version}/{endpoint}" if supports_version else endpoint
        while url:
            response = self.base_request(
                method=method,
                url=url,
                client=client,
                data=data,
                json=json,
                params=params,
                retries=retries,
            )
            if data_key and response:
                data_to_add = response.get(data_key, [])
                if not isinstance(data_to_add, list):
                    data_to_add = [data_to_add]
                results.extend(data_to_add)
            else:
                results.append(response)
            # Check for pagination
            url = response.get(RESPONSE_LINKS_KEY, {}).get(PAGINATION_NEXT) if response else None
        return results

    def converted_request(
        self,
        endpoint: str,
        method: str,
        supports_version: bool = True,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        convert_to_table: bool = True,
        data_key: Optional[str] = None,
        client: Optional[OAuth2APIConnector] = None,
        override_api_version: Optional[str] = None,
    ) -> Union[Table, Dict[str, Any]]:
        """Internal method to make a call to the Newmode API and convert the result to a Parsons table."""

        if params is None:
            params = {}
        client = client if client else self.default_client
        response = self.paginate_request(
            method=method,
            json=json,
            data=data,
            params=params,
            data_key=data_key,
            supports_version=supports_version,
            endpoint=endpoint,
            client=client,
            override_api_version=override_api_version,
        )
        if response:
            if convert_to_table:
                return client.convert_to_table(data=response)
            else:
                return response

    def get_campaign(self, campaign_id: str, params: Optional[Dict[str, Any]] = None) -> Table:
        """
        Retrieve a specific campaign by ID.

        In v2, a campaign is equivalent to Tools or Actions in V1.
        `Args:`
            campaign_id: str
                The ID of the campaign to retrieve.
            params: dict
                Query parameters to include in the request.
        `Returns:`
            Parsons Table containing campaign data.
        """
        if params is None:
            params = {}
        endpoint = f"/campaign/{campaign_id}/form"
        data = self.converted_request(
            endpoint=endpoint,
            method="GET",
            params=params,
        )
        return data

    def get_campaign_ids(self, params: Optional[Dict[str, Any]] = None) -> List[str]:
        """
        Retrieve all campaigns
        In v2, a campaign is equivalent to Tools or Actions in V1.
        `Args:`
            organization_id: str
                ID of organization
            params: dict
                Query parameters to include in the request.
        `Returns:`
            List containing all campaign ids.
        """
        if params is None:
            params = {}
        endpoint = "node/action"
        campaigns_client = OAuth2APIConnector(
            uri=V2_API_CAMPAIGNS_URL,
            auto_refresh_url=V2_API_AUTH_URL,
            client_id=self.client_id,
            client_secret=self.client_secret,
            headers=V2_API_CAMPAIGNS_HEADERS,
            token_url=V2_API_AUTH_URL,
            grant_type="client_credentials",
        )

        data = self.converted_request(
            endpoint=endpoint,
            method="GET",
            params=params,
            data_key=RESPONSE_DATA_KEY,
            client=campaigns_client,
            override_api_version=V2_API_CAMPAIGNS_VERSION,
        )
        return data["id"]

    def get_recipient(
        self,
        campaign_id: str,
        street_address: Optional[str] = None,
        city: Optional[str] = None,
        postal_code: Optional[str] = None,
        region: Optional[str] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Table:
        """
        Retrieve a specific recipient by ID
        `Args:`
            campaign_id: str
                The ID of the campaign to retrieve.
            street_address: str
                Street address of recipient
            city: str
                City of recipient
            postal_code: str
                Postal code of recipient
            region: str
                Region (i.e. state/province abbreviation) of recipient
            params: dict
                Query parameters to include in the request.
        `Returns:`
            Parsons Table containing recipient data.
        """
        if params is None:
            params = {}
        address_params = {
            "street_address": street_address,
            "city": city,
            "postal_code": postal_code,
            "region": region,
        }
        all_address_params_are_missing = all(x is None for x in address_params.values())
        if all_address_params_are_missing:
            raise ValueError(
                "Incomplete Request. Please specify a street address, city, postal code, and/or region."
            )

        params = {f"address[value][{key}]": value for key, value in address_params.items() if value}
        response = self.converted_request(
            endpoint=f"campaign/{campaign_id}/target",
            method="GET",
            params=params,
        )
        return response

    def run_submit(
        self,
        campaign_id: str,
        json: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """
        Pass a submission from a supporter to a campaign
        that ultimately fills in a petition,
        sends an email or triggers a phone call
        depending on your campaign type

        `Args:`
            campaign_id: str
                The ID of the campaign to retrieve.
            params: dict
                Query parameters to include in the request.
        `Returns:`
            Parsons Table containing submit data.
        """

        if params is None:
            params = {}
        response = self.converted_request(
            endpoint=f"campaign/{campaign_id}/submit",
            method="POST",
            data=data,
            json=json,
            params=params,
            convert_to_table=False,
        )
        return response[0]

    def get_submissions(self, campaign_id: str, params: Optional[Dict[str, Any]] = None) -> Table:
        """
        Retrieve and sort submissions and contact data
        for a specified campaign using a range of filters
        that include campaign id, data range and submission status

        `Args:`
            params: dict
                Query parameters to include in the request.
        `Returns:`
            Parsons Table containing submit data.
        """
        if params is None:
            params = {}
        params = {"action": campaign_id}
        response = self.converted_request(
            endpoint="submission", method="GET", params=params, data_key=RESPONSE_DATA_KEY
        )
        return response


class Newmode:
    def __new__(
        cls,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        api_user: Optional[str] = None,
        api_password: Optional[str] = None,
        api_version: str = "v1.0",
    ) -> Union[NewmodeV1, NewmodeV2]:
        """
        Create and return Newmode instance based on chosen version (V1 or V2)

        `Args`:
            api_user: str
                The Newmode api user. Not required if ``NEWMODE_API_USER`` env variable is
                passed. Needed for V1.
            api_password: str
                The Newmode api password. Not required if ``NEWMODE_API_PASSWORD`` env variable is
                passed. Needed for V1.
            client_id: str
                The client id to use for the API requests. Not required if ``NEWMODE_API_CLIENT_ID``
                env variable set. Needed for V2.
            client_secret: str
                The client secret to use for the API requests. Not required if ``NEWMODE_API_CLIENT_SECRET``
                env variable set. Needed for V2.
            api_version: str
                The api version to use. Defaults to v1.0.

        Returns:
            NewMode Class
        """
        api_version = check_env.check("NEWMODE_API_VERSION", api_version)
        if api_version.startswith("v2"):
            return NewmodeV2(
                client_id=client_id, client_secret=client_secret, api_version=api_version
            )
        if api_version.startswith("v1"):
            return NewmodeV1(api_user=api_user, api_password=api_password, api_version=api_version)
        raise ValueError(f"{api_version} not supported.")



import logging

from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector
from parsons.utilities.datetime import convert_unix_to_readable

logger = logging.getLogger(__name__)

EMPOWER_API_ENDPOINT = "https://api.getempower.com/v1/export"


class Empower(object):
    """
    Instantiate class.

        `Args:`
            api_key: str
                The Empower provided API key.The Empower provided Client UUID. Not
                required if ``EMPOWER_API_KEY`` env variable set.
            empower_uri: str
                The URI to access the Empower API. The default is currently set to
                https://api.getempower.com/v1/export. You can set an ``EMPOWER_URI`` env
                variable or use this URI parameter if a different endpoint is necessary.
            cache: boolean
                The Empower API returns all account data after each call. Setting cache
                to ``True`` stores the blob and then extracts Parsons tables for each method.
                Setting cache to ``False`` will download all account data for each method call.
    """

    def __init__(self, api_key=None, empower_uri=None, cache=True):
        self.api_key = check_env.check("EMPOWER_API_KEY", api_key)
        self.empower_uri = (
            check_env.check("EMPOWER_URI", empower_uri, optional=True) or EMPOWER_API_ENDPOINT
        )
        self.headers = {"accept": "application/json", "secret-token": self.api_key}
        self.client = APIConnector(
            self.empower_uri,
            headers=self.headers,
        )
        self.data = None
        self.data = self._get_data(cache)

    def _get_data(self, cache):
        """
        Gets fresh data from Empower API based on cache setting.
        """
        if not cache or self.data is None:
            r = self.client.get_request(self.empower_uri)
            logger.info("Empower data downloaded.")
            return r

        else:
            return self.data

    def _empty_obj(self, obj_name):
        """
        Determine if a dict object is empty.
        """

        if len(self.data[obj_name]) == 0:
            return True
        else:
            return False

    def get_profiles(self):
        """
        Get Empower profiles.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.data["profiles"])
        for col in ["createdMts", "lastUsedEmpowerMts", "updatedMts"]:
            tbl.convert_column(col, lambda x: convert_unix_to_readable(x))
        tbl.remove_column("activeCtaIds")  # Get as a method via get_profiles_active_ctas
        return tbl

    def get_profiles_active_ctas(self):
        """
        Get active ctas assigned to Empower profiles.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.data["profiles"]).long_table("eid", "activeCtaIds")
        return tbl

    def get_regions(self):
        """
        Get Empower regions.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.data["regions"])
        tbl.convert_column("inviteCodeCreatedMts", lambda x: convert_unix_to_readable(x))
        return tbl

    def get_cta_results(self):
        """
        Get Empower call to action results.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        # unpacks answerIdsByPromptId into standalone rows
        tbl = Table(self.data["ctaResults"])
        tbl.convert_column("contactedMts", lambda x: convert_unix_to_readable(x))
        tbl = tbl.unpack_nested_columns_as_rows(
            "answerIdsByPromptId", key="profileEid", expand_original=True
        )
        tbl.unpack_list("answerIdsByPromptId_value", replace=True)
        col_list = [v for v in tbl.columns if v.find("value") != -1]
        tbl.coalesce_columns("answer_id", col_list, remove_source_columns=True)
        tbl.remove_column("uid")
        tbl.remove_column("answers")  # Per docs, this is deprecated.
        return tbl

    def _split_ctas(self):
        """
        Internal method to split CTA objects into tables.
        """

        ctas = Table(self.data["ctas"])
        for col in [
            "createdMts",
            "scheduledLaunchTimeMts",
            "updatedMts",
            "activeUntilMts",
        ]:
            ctas.convert_column(col, lambda x: convert_unix_to_readable(x))
        # Get following data as their own tables via their own methods
        ctas.remove_column("regionIds")  # get_cta_regions()
        ctas.remove_column("shareables")  # get_cta_shareables()
        ctas.remove_column("prioritizations")  # get_cta_prioritizations()
        ctas.remove_column("questions")  # This column has been deprecated.

        cta_prompts = ctas.long_table("id", "prompts", prepend=False, retain_original=False)
        cta_prompts.remove_column("ctaId")

        cta_prompt_answers = cta_prompts.long_table("id", "answers", prepend=False)

        return {
            "ctas": ctas,
            "cta_prompts": cta_prompts,
            "cta_prompt_answers": cta_prompt_answers,
        }

    def get_ctas(self):
        """
        Get Empower calls to action.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        return self._split_ctas()["ctas"]

    def get_cta_prompts(self):
        """
        Get Empower calls to action prompts.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        return self._split_ctas()["cta_prompts"]

    def get_cta_prompt_answers(self):
        """
        Get Empower calls to action prompt answers.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        return self._split_ctas()["cta_prompt_answers"]

    def get_cta_regions(self):
        """
        Get a list of regions that each call to active is active in.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.data["ctas"]).long_table("id", "regionIds")
        return tbl

    def get_cta_shareables(self):
        """
        Get a list of shareables associated with calls to action.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.data["ctas"]).long_table("id", "shareables")
        return tbl

    def get_cta_prioritizations(self):
        """
        Get a list prioritizations associated with calls to action.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.data["ctas"]).long_table("id", "prioritizations")
        return tbl

    def get_outreach_entries(self):
        """
        Get outreach entries.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        if self._empty_obj("outreachEntries"):
            logger.info("No Outreach Entries found.")
            return Table([])

        tbl = Table(self.data["outreachEntries"])
        for col in [
            "outreachCreatedMts",
            "outreachSnoozeUntilMts",
            "outreachScheduledFollowUpMts",
        ]:
            tbl.convert_column(col, lambda x: convert_unix_to_readable(x))
        return tbl

    def get_full_export(self):
        """
        Get a table of the complete, raw data as returned by the API.
        Meant to facilitate pure ELT pipelines

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table([self.data])
        return tbl



import json
import logging

import requests

from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

URI = "https://api.bloomerang.co/v2/"
URI_AUTH = "https://crm.bloomerang.co/authorize/"


class Bloomerang(object):
    """
    Instantiate Bloomerang class

    `Args:`
        api_key: str
            The Bloomerang API key. Not required if the ``BLOOMERANG_API_KEY`` environmental
            variable is set or if the OAuth2 authentication parameters ``client_id`` and
            ``client_secret`` are set.
        client_id: str
            The Bloomerang client ID for OAuth2 authentication. Not required if
            the ``BLOOMERANG_CLIENT_ID`` env variable is set or if the ``api_key``
            parameter is set. Note that the ``client_secret`` parameter must
            also be set in order to use OAuth2 authentication.
        client_secret: str
            The Bloomerang client secret for OAuth2 authetication. Not required if
            the ``BLOOMERANG_CLIENT_SECRET`` env variable is set or if the ``api_key``
            parameter is set. Note that the ``client_id`` parameter must
            also be set in order to use OAuth2 authentication.
    """

    def __init__(self, api_key=None, client_id=None, client_secret=None):
        self.api_key = check_env.check("BLOOMERANG_API_KEY", api_key, optional=True)
        self.client_id = check_env.check("BLOOMERANG_CLIENT_ID", client_id, optional=True)
        self.client_secret = check_env.check(
            "BLOOMERANG_CLIENT_SECRET", client_secret, optional=True
        )
        self.uri = URI
        self.uri_auth = URI_AUTH
        self.conn = self._conn()

    def _conn(self):
        # Instantiate APIConnector with authentication credentials
        headers = {"accept": "application/json", "Content-Type": "application/json"}
        if self.api_key is not None:
            logger.info("Using API key authentication.")
            headers["X-API-KEY"] = f"{self.api_key}"
        elif (self.client_id is not None) & (self.client_secret is not None):
            logger.info("Using OAuth2 authentication.")
            self._generate_authorization_code()
            self._generate_access_token()
            headers["Authorization"] = f"Bearer {self.access_token}"
        else:
            raise Exception("Missing authorization credentials.")
        return APIConnector(uri=self.uri, headers=headers)

    def _generate_authorization_code(self):
        data = {"client_id": self.client_id, "response_type": "code"}
        r = requests.post(url=self.uri_auth, json=data)
        self.authorization_code = r.json().get("code", None)

    def _generate_access_token(self):
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "authorization_code",
            "code": self.authorization_code,
        }
        r = requests.post(url=self.uri + "oauth/token", json=data)
        self.access_token = r.json().get("access_token", None)

    def _base_endpoint(self, endpoint, entity_id=None):
        url = f"{self.uri}{endpoint}/"
        if entity_id:
            url = url + f"{entity_id}/"
        return url

    @staticmethod
    def _base_pagination_params(page_number=1, page_size=50):
        return {"skip": page_size * (page_number - 1), "take": min(page_size, 50)}

    @staticmethod
    def _base_ordering_params(order_by=None, order_direction=None):
        params = {}

        if order_by:
            params["orderBy"] = order_by

        if order_direction:
            params["orderDirection"] = order_direction

        return params

    def _base_create(self, endpoint, entity_id=None, **kwargs):
        return self.conn.post_request(
            url=self._base_endpoint(endpoint, entity_id), json=json.dumps({**kwargs})
        )

    def _base_update(self, endpoint, entity_id=None, **kwargs):
        return self.conn.put_request(
            url=self._base_endpoint(endpoint, entity_id), json=json.dumps({**kwargs})
        )

    def _base_get(self, endpoint, entity_id=None, params=None):
        return self.conn.get_request(url=self._base_endpoint(endpoint, entity_id), params=params)

    def _base_delete(self, endpoint, entity_id=None):
        return self.conn.delete_request(url=self._base_endpoint(endpoint, entity_id))

    def create_constituent(self, **kwargs):
        """
        `Args:`
            **kwargs:`
                Fields to include, e.g., FirstName = 'Rachel'.

                See the Bloomerang API docs for a full list of `fields <https://bloomerang.co/features/integrations/api/rest-api#/Constituents/post_constituent>`_.
        """
        return self._base_create("constituent", **kwargs)

    def update_constituent(self, constituent_id, **kwargs):
        """
        `Args:`
            constituent_id: str or int
                Constituent ID to update
            **kwargs:`
                Fields to update, e.g., FirstName = 'RJ'.

                See the Bloomerang API docs for a full list of `fields <https://bloomerang.co/features/integrations/api/rest-api#/Constituents/put_constituent__id_>`_.
        """
        return self._base_update("constituent", entity_id=constituent_id, **kwargs)

    def get_constituent(self, constituent_id):
        """
        `Args:`
            constituent_id: str or int
                Constituent ID to get fields for
        `Returns:`
            A  JSON of the entry or an error.
        """
        return self._base_get("constituent", entity_id=constituent_id)

    def delete_constituent(self, constituent_id):
        """
        `Args:`
            constituent_id: str or int
                Constituent ID to delete
        """
        return self._base_delete("constituent", entity_id=constituent_id)

    def get_constituents(
        self,
        page_number=1,
        page_size=50,
        order_by=None,
        order_direction=None,
        last_modified=None,
    ):
        """
        `Args:`
            page_number: int
                Number of the page to fetch
            page_size: int
                Number of records per page (maximum allowed is 50)
            order_by: str
                Sorts by ``Id``, ``CreatedDate``, or ``LastModifiedDate`` (default ``Id``).
            order_direction: str
                Sorts the order_by in ``Asc`` or ``Desc`` order.
            last_modified: str
                Filters to constituents last modified after the specified date (ISO-8601 format).
        `Returns:`
            A Table of the entries.
        """
        params = self._base_pagination_params(page_number, page_size)
        params.update(self._base_ordering_params(order_by, order_direction))

        if last_modified:
            params["lastModified"] = last_modified

        response = self._base_get("constituents", params=params)
        return Table(response["Results"])

    def create_transaction(self, **kwargs):
        """
        `Args:`
            **kwargs:`
                Fields to include, e.g., CreditCardType = 'Visa'.

                See the Bloomerang API docs for a full list of `fields <https://bloomerang.co/features/integrations/api/rest-api#/Transactions/post_transaction>`_.
        """
        return self._base_create("transaction", **kwargs)

    def update_transaction(self, transaction_id, **kwargs):
        """
        `Args:`
            transaction_id: str or int
                Transaction ID to update
            **kwargs:`
                Fields to update, e.g., CreditCardType = 'Visa'.

                See the Bloomerang API docs for a full list of `fields <https://bloomerang.co/features/integrations/api/rest-api#/Transactions/put_transaction__id_>`_.
        """
        return self._base_update("transaction", entity_id=transaction_id, **kwargs)

    def get_transaction(self, transaction_id):
        """
        `Args:`
            transaction_id: str or int
                Transaction ID to get fields for
        `Returns:`
            A  JSON of the entry or an error.
        """
        return self._base_get("transaction", entity_id=transaction_id)

    def delete_transaction(self, transaction_id):
        """
        `Args:`
            transaction_id: str or int
                Transaction ID to delete
        """
        return self._base_delete("transaction", entity_id=transaction_id)

    def get_transactions(self, page_number=1, page_size=50, order_by=None, order_direction=None):
        """
        `Args:`
            page_number: int
                Number of the page to fetch
            page_size: int
                Number of records per page (maximum allowed is 50)
            order_by: str
                Sorts by ``Date``, ``CreatedDate``, or ``LastModifiedDate`` (default ``Date``).
            order_direction: str
                Sorts the order_by in ``Asc`` or ``Desc`` order (default ``Desc``).
        `Returns:`
            A  JSON of the entry or an error.
        """
        params = self._base_pagination_params(page_number, page_size)
        params.update(self._base_ordering_params(order_by, order_direction))

        response = self._base_get("transactions", params=params)
        return Table(response["Results"])

    def get_transaction_designation(self, designation_id):
        """
        `Args:`
            designation_id: str or int
                Transaction Designation ID to get fields for
        `Returns:`
            A  JSON of the entry or an error.
        """
        return self._base_get("transaction/designation", entity_id=designation_id)

    def get_transaction_designations(
        self, page_number=1, page_size=50, order_by=None, order_direction=None
    ):
        """
        `Args:`
            page_number: int
                Number of the page to fetch
            page_size: int
                Number of records per page (maximum allowed is 50)
            order_by: str
                Sorts by ``Date``, ``CreatedDate``, or ``LastModifiedDate`` (default ``Date``).
            order_direction: str
                Sorts the order_by in ``Asc`` or ``Desc`` order (default ``Desc``).
        `Returns:`
            A  JSON of the entry or an error.
        """
        params = self._base_pagination_params(page_number, page_size)
        params.update(self._base_ordering_params(order_by, order_direction))

        response = self._base_get("transactions/designations", params=params)
        return Table(response["Results"])

    def create_interaction(self, **kwargs):
        """
        `Args:`
            **kwargs:`
                Fields to include, e.g., Channel = "Email".

                See the Bloomerang API docs for a full list of `fields <https://bloomerang.co/features/integrations/api/rest-api#/Interactions/post_interaction>`_.
        """
        return self._base_create("interaction", **kwargs)

    def update_interaction(self, interaction_id, **kwargs):
        """
        `Args:`
            interaction_id: str or int
                Interaction ID to update
            **kwargs:`
                Fields to update, e.g., EmailAddress = "user@example.com".

                See the Bloomerang API docs for a full list of `fields <https://bloomerang.co/features/integrations/api/rest-api#/Interactions/put_interaction__id_>`_.
        """
        return self._base_update("interaction", entity_id=interaction_id, **kwargs)

    def get_interaction(self, interaction_id):
        """
        `Args:`
            interaction_id: str or int
                Interaction ID to get fields for
        `Returns:`
            A  JSON of the entry or an error.
        """
        return self._base_get("interaction", entity_id=interaction_id)

    def delete_interaction(self, interaction_id):
        """
        `Args:`
            interaction_id: str or int
                Interaction ID to delete
        """
        return self._base_delete("interaction", entity_id=interaction_id)

    def get_interactions(self, page_number=1, page_size=50):
        """
        `Args:`
            page_number: int
                Number of the page to fetch
            page_size: int
                Number of records per page (maximum allowed is 50)
        `Returns:`
            A  JSON of the entry or an error.
        """
        params = self._base_pagination_params(page_number, page_size)
        response = self._base_get("interactions", params=params)
        return Table(response["Results"])



import requests

from parsons.etl import Table
from parsons.utilities import check_env

URI = "https://www.googleapis.com/civicinfo/v2/"


class GoogleCivic(object):
    """
    `Args:`
        api_key : str
            A valid Google api key. Not required if ``GOOGLE_CIVIC_API_KEY``
            env variable set.
    `Returns:`
        class
    """

    def __init__(self, api_key=None):
        self.api_key = check_env.check("GOOGLE_CIVIC_API_KEY", api_key)
        self.uri = URI

    def request(self, url, args=None):
        # Internal request method

        if not args:
            args = {}

        args["key"] = self.api_key

        r = requests.get(url, params=args)

        return r.json()

    def get_elections(self):
        """
        Get a collection of information about elections and voter information.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        url = self.uri + "elections"

        return Table((self.request(url))["elections"])

    def _get_voter_info(self, election_id, address):
        # Internal method to call voter info end point. Portions of this are
        # parsed for other methods.

        url = self.uri + "voterinfo"

        args = {"address": address, "electionId": election_id}

        return self.request(url, args=args)

    def get_polling_location(self, election_id, address):
        """
        Get polling location information for a given address.

        `Args:`
            election_id: int
                A valid election id. Election ids can be found by running the
                :meth:`get_elections` method.
            address: str
                A valid US address in a single string.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self._get_voter_info(election_id, address)

        return r["pollingLocations"]

    def get_polling_locations(self, election_id, table, address_field="address"):
        """
        Get polling location information for a table of addresses.

        `Args:`
            election_id: int
                A valid election id. Election ids can be found by running the
                :meth:`get_elections` method.
            address: str
                A valid US address in a single string.
            address_field: str
                The name of the column where the address is stored.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        polling_locations = []

        # Iterate through the rows of the table
        for row in table:
            loc = self.get_polling_location(election_id, row[address_field])
            # Insert original passed address
            loc[0]["passed_address"] = row[address_field]

            # Add to list of lists
            polling_locations.append(loc[0])

        # Unpack values
        tbl = Table(polling_locations)
        tbl.unpack_dict("address", prepend_value="polling")
        tbl.unpack_list("sources", replace=True)
        tbl.unpack_dict("sources_0", prepend_value="source")
        tbl.rename_column("polling_line1", "polling_address")

        # Resort columns
        tbl.move_column("pollingHours", len(tbl.columns))
        tbl.move_column("notes", len(tbl.columns))
        tbl.move_column("polling_locationName", 1)
        tbl.move_column("polling_address", 2)

        return tbl

    def get_representative_info_by_address(
        self, address: str, include_offices=True, levels=None, roles=None
    ):
        """
        Get representative information for a given address.
        This method returns the raw JSON response from the Google Civic API.
        It is a complex response that is not easily parsed into a table.
        Here is the information on how to parse the response:
        https://developers.google.com/civic-information/docs/v2/representatives/representativeInfoByAddress

        `Args:`
            address: str
                A valid US address in a single string.
            include_offices: bool
                Whether to return information about offices and officials.
                If false, only the top-level district information will be returned. (Default: True)
            levels: list of str
                A list of office levels to filter by.
                Only offices that serve at least one of these levels will be returned.
                Divisions that don't contain a matching office will not be returned.
                    Acceptable values are:
                    "administrativeArea1"
                    "administrativeArea2"
                    "country"
                    "international"
                    "locality"
                    "regional"
                    "special"
                    "subLocality1"
                    "subLocality2"
            roles: list of str
                A list of office roles to filter by.
                Only offices fulfilling one of these roles will be returned.
                Divisions that don't contain a matching office will not be returned.
                    Acceptable values are:
                    "deputyHeadOfGovernment"
                    "executiveCouncil"
                    "governmentOfficer"
                    "headOfGovernment"
                    "headOfState"
                    "highestCourtJudge"
                    "judge"
                    "legislatorLowerBody"
                    "legislatorUpperBody"
                    "schoolBoard"
                    "specialPurposeOfficer"

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if levels is not None and not isinstance(levels, list):
            raise ValueError("levels must be a list of strings")
        if roles is not None and not isinstance(roles, list):
            raise ValueError("roles must be a list of strings")
        if address is None or not isinstance(address, str):
            raise ValueError("address must be a string")

        url = self.uri + "representatives"

        args = {
            "address": address,
            "includeOffices": include_offices,
            "levels": levels,
            "roles": roles,
        }

        response = self.request(url, args=args)

        # Raise an error if the address was invalid
        if "error" in response:
            raise ValueError(response["error"]["message"])

        return response



import logging
from datetime import datetime, timedelta
from typing import Dict, NoReturn, Optional, Union

from requests import Response, request

from parsons.etl import Table
from parsons.hustle.column_map import LEAD_COLUMN_MAP
from parsons.utilities import check_env, json_format

logger = logging.getLogger(__name__)

HUSTLE_URI = "https://api.hustle.com/v1/"
PAGE_LIMIT = 1000


class Hustle(object):
    """
    Instantiate Hustle Class

    `Args:`
        client_id:
            The client id provided by Hustle. Not required if ``HUSTLE_CLIENT_ID`` env variable
            set.
        client_secret:
            The client secret provided by Hustle. Not required if ``HUSTLE_CLIENT_SECRET`` env
            variable set.
    `Returns:`
        Hustle Class
    """

    def __init__(self, client_id: Optional[str] = None, client_secret: Optional[str] = None):
        self.uri = HUSTLE_URI
        self.client_id = check_env.check("HUSTLE_CLIENT_ID", client_id)
        self.client_secret = check_env.check("HUSTLE_CLIENT_SECRET", client_secret)
        self.auth_token, self.token_expiration = self._get_auth_token(
            self.client_id, self.client_secret
        )

    def _get_auth_token(self, client_id: str, client_secret: str):
        """Generate an authorization token."""

        data = {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "client_credentials",
        }

        resp = request("POST", self.uri + "oauth/token", data=data)
        resp_json = resp.json()
        logger.debug(resp_json)

        auth_token = resp_json["access_token"]
        token_expiration = datetime.now() + timedelta(seconds=resp_json["expires_in"])
        logger.info("Authentication token generated")
        return auth_token, token_expiration

    def _refresh_token(self):
        """Generate new token if current token is exprired.

        Tokens are valid for `expires_in` (7200 by default) seconds.
        """

        logger.debug("Checking token expiration.")

        if datetime.now() >= self.token_expiration:
            logger.info("Refreshing authentication token.")
            self.auth_token, self.token_expiration = self._get_auth_token(
                self.client_id, self.client_secret
            )

    def _request(
        self,
        endpoint: str,
        req_type: str = "GET",
        args: Optional[Dict] = None,
        payload: Optional[Dict] = None,
        raise_on_error: bool = True,
    ) -> Union[Dict, list]:
        url = self.uri + endpoint
        self._refresh_token()

        headers = {"Authorization": f"Bearer {self.auth_token}"}

        parameters = {}
        if req_type == "GET":
            parameters = {"limit": PAGE_LIMIT}

        if args:
            parameters.update(args)

        resp = request(req_type, url, params=parameters, json=payload, headers=headers)

        self._error_check(resp, raise_on_error)
        resp_json = resp.json()

        # If a single item return the dict
        if "items" not in resp_json.keys():
            return resp_json

        result = resp_json["items"]

        # Pagination
        while resp_json["pagination"]["hasNextPage"] == "true":
            parameters["cursor"] = resp_json["pagination"]["cursor"]
            resp = request(req_type, url, params=parameters, headers=headers)
            self._error_check(resp, raise_on_error)
            resp_json = resp.json()
            result += resp_json["items"]

        return result

    def _error_check(self, resp: Response, raise_on_error: bool) -> Optional[NoReturn]:
        """Check response for errors."""

        if resp.status_code in (200, 201):
            logger.debug(resp.json())
            return

        if raise_on_error:
            logger.info(resp.json())
            resp.raise_for_status()
            return

        logger.info(resp.json())
        return

    def get_agents(self, group_id: str) -> Table:
        """
        Get a list of agents.

        `Args:`
            group_id: str
                The group id.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request(f"groups/{group_id}/agents"))
        logger.info(f"Got {tbl.num_rows} agents from {group_id} group.")
        return tbl

    def get_agent(self, agent_id: str) -> Dict:
        """
        Get a single agent.

        `Args:`
            agent_id: str
                The agent id.
        `Returns:`
            dict
        """

        resp = self._request(f"agents/{agent_id}")
        logger.info(f"Got {agent_id} agent.")
        return resp  # type: ignore

    def create_agent(
        self,
        group_id: str,
        name: str,
        full_name: str,
        phone_number: str,
        send_invite: bool = False,
        email: Optional[str] = None,
    ) -> Dict:
        """
        Create an agent.

        `Args:`
            group_id: str
                The group id to assign the agent.
            name: str
                The name of the agent.
            full_name: str
                The full name of the agent.
            phone_number: str
                The valid phone number of the agent.
            send_invite: boolean
                Send an invitation to the agent.
            email:
                The email address of the agent.
        `Returns:`
            dict
        """

        agent = {
            "name": name,
            "fullName": full_name,
            "phoneNumber": phone_number,
            "sendInvite": send_invite,
            "email": email,
        }

        # Remove empty args in dictionary
        agent = json_format.remove_empty_keys(agent)

        logger.info(f"Generating {full_name} agent.")
        resp = self._request(f"groups/{group_id}/agents", req_type="POST", payload=agent)
        return resp  # type: ignore

    def update_agent(
        self,
        agent_id: str,
        name: Optional[str] = None,
        full_name: Optional[str] = None,
        send_invite: bool = False,
    ) -> Dict:
        """
        Update an agent.

        `Args:`
            agent_id: str
                The agent id.
            name: str
                The name of the agent.
            full_name: str
                The full name of the agent.
            phone_number: str
                The valid phone number of the agent.
            send_invite: boolean
                Send an invitation to the agent.
        `Returns:`
            dict
        """

        agent = {"name": name, "fullName": full_name, "sendInvite": send_invite}

        # Remove empty args in dictionary
        agent = json_format.remove_empty_keys(agent)

        logger.info(f"Updating agent {agent_id}.")
        resp = self._request(f"agents/{agent_id}", req_type="PUT", payload=agent)
        return resp  # type: ignore

    def get_organizations(self) -> Table:
        """
        Get organizations.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request("organizations"))
        logger.info(f"Got {tbl.num_rows} organizations.")
        return tbl

    def get_organization(self, organization_id: str) -> Dict:
        """
        Get a single organization.

        `Args:`
            organization_id: str
                The organization id.
        `Returns:`
            dict
        """

        resp = self._request(f"organizations/{organization_id}")
        logger.info(f"Got {organization_id} organization.")
        return resp  # type: ignore

    def get_groups(self, organization_id: str) -> Table:
        """
        Get a list of groups.

        `Args:`
            organization_id: str
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request(f"organizations/{organization_id}/groups"))
        logger.info(f"Got {tbl.num_rows} groups.")
        return tbl

    def get_group(self, group_id: str) -> Dict:
        """
        Get a single group.

        `Args:`
            group_id: str
                The group id.
        """

        resp = self._request(f"groups/{group_id}")
        logger.info(f"Got {group_id} group.")
        return resp  # type: ignore

    def create_group_membership(self, group_id: str, lead_id: str) -> Dict:
        """
        Add a lead to a group.

        `Args:`
            group_id: str
                The group id.
            lead_id: str
                The lead id.
        """

        resp = self._request(
            f"groups/{group_id}/memberships",
            req_type="POST",
            payload={"leadId": lead_id},
        )
        return resp  # type: ignore

    def get_lead(self, lead_id: str) -> Dict:
        """
        Get a single lead.

        `Args`:
            lead_id: str
                The lead id.
        `Returns:`
            dict
        """

        resp = self._request(f"leads/{lead_id}")
        logger.info(f"Got {lead_id} lead.")
        return resp  # type: ignore

    def get_leads(
        self, organization_id: Optional[str] = None, group_id: Optional[str] = None
    ) -> Table:
        """
        Get leads metadata. One of ``organization_id`` and ``group_id`` must be passed
        as an argument. If both are passed, an error will be raised.

        `Args:`
            organization_id: str
                The organization id.
            group_id: str
                The group id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if organization_id is None and group_id is None:
            raise ValueError("Either organization_id or group_id required.")

        if organization_id is not None and group_id is not None:
            raise ValueError("Only one of organization_id and group_id may be populated.")

        if organization_id:
            endpoint = f"organizations/{organization_id}/leads"
            logger.info(f"Retrieving {organization_id} organization leads.")
        if group_id:
            endpoint = f"groups/{group_id}/leads"
            logger.info(f"Retrieving {group_id} group leads.")

        tbl = Table(self._request(endpoint))  # type: ignore
        logger.info(f"Got {tbl.num_rows} leads.")
        return tbl

    def create_lead(
        self,
        group_id: str,
        phone_number: str,
        first_name: str,
        last_name: Optional[str] = None,
        email: Optional[str] = None,
        notes: Optional[str] = None,
        follow_up: Optional[str] = None,
        custom_fields: Optional[Dict] = None,
        tag_ids: Optional[list] = None,
    ) -> Dict:
        """

        Create a lead.

        `Args:`
            group_id: str
                The group id to assign the leads.
            first_name: str
                The first name of the lead.
            phone_number: str
                The phone number of the lead.
            last_name: str
                The last name of the lead.
            email: str
                The email address of the lead.
            notes: str
                The notes for the lead.
            follow_up: str
                Follow up for the lead.
            custom_fields: dict
                A dictionary of custom fields, with key as the value name, and
                value as the value.
            tag_ids: list
                A list of tag ids.
        `Returns:`
                ``None``
        """

        lead = {
            "firstName": first_name,
            "lastName": last_name,
            "email": email,
            "phoneNumber": phone_number,
            "notes": notes,
            "followUp": follow_up,
            "customFields": custom_fields,
            "tagIds": tag_ids,
        }

        # Remove empty args in dictionary
        lead = json_format.remove_empty_keys(lead)
        logger.info(f"Generating lead for {first_name} {last_name}.")
        resp = self._request(f"groups/{group_id}/leads", req_type="POST", payload=lead)
        return resp  # type: ignore

    def create_leads(self, table: Table, group_id: Optional[str] = None) -> Table:
        """
        Create multiple leads. All unrecognized fields will be passed as custom fields. Column
        names must map to the following names.

        .. list-table::
            :widths: 20 80
            :header-rows: 1

            * - Column Name
              - Valid Column Names
            * - first_name
              - ``first_name``, ``first``, ``fn``, ``firstname``
            * - last_name
              - ``last_name``, ``last``, ``ln``, ``lastname``
            * - phone_number
               - ``phone_number``, ``phone``, ``cell``, ``phonenumber``, ``cell_phone``
                ``cellphone``
            * - email
              - ``email``, ``email_address``, ``emailaddress``
            * - follow_up
              - ``follow_up``, ``followup``

        `Args:`
            table: Parsons table
                A Parsons table containing leads
            group_id:
                The group id to assign the leads. If ``None``, must be passed as a column
                value.
        `Returns:`
            A table of created ids with associated lead id.
        """

        table.map_columns(LEAD_COLUMN_MAP)

        arg_list = [
            "first_name",
            "last_name",
            "email",
            "phone_number",
            "follow_up",
            "tag_ids",
            "group_id",
        ]

        created_leads = []

        for row in table:
            lead: Dict[str, Optional[Union[str, Dict]]] = {"group_id": group_id}
            custom_fields = {}

            # Check for column names that map to arguments, if not assign
            # to custom fields
            for k, v in row.items():
                if k in arg_list:
                    lead[k] = v
                else:
                    custom_fields[k] = v

            lead["custom_fields"] = custom_fields

            # Group Id check
            if not group_id and "group_id" not in table.columns:
                raise ValueError("Group Id must be passed as an argument or a column value.")
            if group_id:
                lead["group_id"] = group_id

            created_leads.append(self.create_lead(**lead))  # type: ignore

        logger.info(f"Created {table.num_rows} leads.")
        return Table(created_leads)

    def update_lead(
        self,
        lead_id: str,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        email: Optional[str] = None,
        global_opt_out: Optional[bool] = None,
        notes: Optional[str] = None,
        follow_up: Optional[str] = None,
        tag_ids: Optional[list] = None,
    ) -> Dict:
        """
        Update a lead.

        `Args`:
            lead_id: str
                The lead id
            first_name: str
                The first name of the lead
            last_name: str
                The last name of the lead
            email: str
                The email address of the lead
            global_opt_out: boolean
                Opt out flag for the lead
            notes: str
                The notes for the lead
            follow_up: str
                Follow up for the lead
            tag_ids: list
                Tags to apply to lead
        `Returns:`
            dict
        """

        lead = {
            "leadId": lead_id,
            "firstName": first_name,
            "lastName": last_name,
            "email": email,
            "globalOptedOut": global_opt_out,
            "notes": notes,
            "followUp": follow_up,
            "tagIds": tag_ids,
        }

        # Remove empty args in dictionary
        lead = json_format.remove_empty_keys(lead)

        logger.info(f"Updating lead for {first_name} {last_name}.")
        resp = self._request(f"leads/{lead_id}", req_type="PUT", payload=lead)
        return resp  # type: ignore

    def get_tags(self, organization_id: str) -> Table:
        """
        Get an organization's tags.

        `Args:`
            organization_id: str
                The organization id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request(f"organizations/{organization_id}/tags"))
        logger.info(f"Got {tbl.num_rows} tags for {organization_id} organization.")
        return tbl

    def get_tag(self, tag_id: str) -> Dict:
        """
        Get a single tag.

        `Args:`
            tag_id: str
                The tag id.
        `Returns:`
            dict
        """

        resp = self._request(f"tags/{tag_id}")
        logger.info(f"Got {tag_id} tag.")
        return resp  # type: ignore

    def get_custom_fields(self, organization_id: str) -> Table:
        """Retrieve an organization's custom fields.

        `Args:`
            organization_id: str
                The organization id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self._request(f"organizations/{organization_id}/custom-fields"))
        logger.info(f"Got {tbl.num_rows} custom fields for {organization_id} organization.")
        return tbl

    def create_custom_field(
        self, organization_id: str, name: str, agent_visible: Optional[bool] = None
    ) -> Dict:
        """Create a custom field.

        `Args:`
            organization_id: str
                The organization id.
            name: str
                The name of the custom field. Restricted to letters, numbers, and underscores. Minimum of 2 characters, maximum of 40.
            agent_visible: bool
                Optional. `true` represents that the custom field is visible to agents. `false` means that only admins can see it.
        `Returns:`
            dict
                The newly created custom field
        """

        custom_field: Dict[str, Union[str, bool]] = {"name": name}
        if agent_visible is not None:
            custom_field["agentVisible"] = agent_visible

        logger.info(f"Generating custom field {name} for organization {organization_id}.")
        resp = self._request(
            f"organizations/{organization_id}/custom-fields", req_type="POST", payload=custom_field
        )
        return resp  # type: ignore



import unittest

from parsons import Table
from parsons.aws.aws_async import get_func_task_path, import_and_get_task
from parsons.aws.lambda_distribute import distribute_task
from test.utils import assert_matching_tables


def fake_func(arg1, arg2, kwarg1=None, kwarg2=56):
    return (arg1, arg2, kwarg1, kwarg2)


count = 0
tableargs = (None, None)


def fake_table_process(table, **fakekwargs):
    global count
    global tableargs
    count = count + 1
    tableargs = (table, fakekwargs)


class FakeRunner(object):
    def __init__(self, init1=None):
        self.init1 = init1

    def foobar(self, table, a, x=None, y=None):
        global count
        global tableargs
        count = count + 1
        tableargs = (table, self.init1, a, x, y)
        if a == "raise":
            raise Exception("foo bar")


class TestAsync(unittest.TestCase):
    def test_task_path_conversion(self):
        # fake_func
        fake_str = get_func_task_path(fake_func)
        print("fake_str", fake_str)
        fake_renewed = import_and_get_task(fake_str)
        self.assertEqual(fake_renewed(1, 2, 3), (1, 2, 3, 56))

        # Table.from_csv_string @classmethod
        csv_str = get_func_task_path(Table.from_csv_string, Table)
        print("csv_str", csv_str)
        csv_renewed = import_and_get_task(csv_str)
        assert_matching_tables(csv_renewed("x,y\n1,2"), Table([("x", "y"), ("1", "2")]))

        # Table.to_dicts (instance)
        dicts_str = get_func_task_path(Table.to_dicts, Table)
        print("dicts_str", dicts_str)
        dicts_renewed = import_and_get_task(dicts_str, {"lst": [("x", "y"), (1, 2)]})
        self.assertEqual(dicts_renewed(), [{"x": 1, "y": 2}])

    def test_distribute_task(self):
        global count
        global tableargs
        datatable = [
            ("x", "y"),
            (1, 2),
            (3, 4),
            (5, 6),
            (7, 8),
            (9, 0),
            (11, 12),
            (13, 14),
            (15, 16),
            (17, 18),
            (19, 10),
        ]
        count = 0
        tableargs = None
        distribute_task(
            Table(datatable),
            fake_table_process,
            "foo",  # bucket
            group_count=5,
            storage="local",
            func_kwargs={"x": 1, "y": [2, 3]},
        )
        self.assertEqual(count, 2)
        assert_matching_tables(
            tableargs[0],
            Table(
                [
                    ("x", "y"),
                    ("11", "12"),
                    ("13", "14"),
                    ("15", "16"),
                    ("17", "18"),
                    ("19", "10"),
                ]
            ),
        )
        count = 0
        tableargs = None
        distribute_task(
            Table(datatable + [(0, 0)]),
            FakeRunner.foobar,
            "foo",  # bucket
            group_count=5,
            storage="local",
            func_class=FakeRunner,
            func_class_kwargs={"init1": "initx"},
            func_kwargs={"a": 1, "x": 2, "y": 3},
        )
        self.assertEqual(count, 3)
        self.assertEqual(tableargs[1:], ("initx", 1, 2, 3))
        self.assertEqual(tableargs[1:], ("initx", 1, 2, 3))
        assert_matching_tables(tableargs[0], Table([("x", "y"), ("0", "0")]))

        # 3. catch=True (with throwing)
        count = 0
        tableargs = None
        distribute_task(
            Table(datatable[:6]),
            FakeRunner.foobar,
            "foo",  # bucket
            group_count=5,
            storage="local",
            func_class=FakeRunner,
            func_class_kwargs={"init1": "initx"},
            catch=True,
            func_kwargs={"a": "raise", "x": 2, "y": 3},
        )



import logging
from typing import Optional

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

API_URI = "https://www.formstack.com/api/v2"


class Formstack(object):
    """
    Instantiate Formstack class.

       `Args:`
            api_token:
                API token to access the Formstack API. Not required if the
                ``FORMSTACK_API_TOKEN`` env variable is set.
    """

    def __init__(self, api_token: Optional[str] = None):
        self.api_token = check_env.check("FORMSTACK_API_TOKEN", api_token)
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.api_token}",
        }
        self.client = APIConnector(API_URI, headers=headers)

    def _get_paginated_request(
        self, url: str, data_key: str, params: dict = None, large_request: bool = False
    ) -> Table:
        """
        Make a GET request for any endpoint that returns a list of data. Will check pagination.

        This is suitable for endpoints like `GET /form`, which gets all forms. It is not suitable
        for endpoints like `GET /form/:id`, which returns the data for one particular form,
        and is never paginated.

        Author's note: I have observed some inconsistency in how Formstack is paginating its
            its endpoints. For example, the /form endpoint (mentioned above) is documented as
            if it follows Formstack's standard pagination scheme. However, it seems to just
            return all of your form data and does not include the pagination keys at all.
            So for developing this connector, I'm just experimenting with endpoints as I write
            the methods for them and using `_get_paginated_request` as needed. Fortunately using
            the connector will remove the need to worry about this from the caller, and we can
            udated the methods in this class as needed when/if Formstack changes pagination on
            their API.

        `Args:`
            url: string
                Relative URL (from the Formstack base URL) to make the request.

            data_key: string
                JSON key that will hold the data in the response body.

            params: Dictionary, optional
                Params to pass to the request.

            large_request: Boolean, optional
                If the response is likely to include a large number of pages. Defaults to `False`.
                In rare cases the API will return more pages than `parsons.Table` is able to handle.
                Pass `True` to enable a workaround for these endpoints.

        `Returns:`
            Table Class
                A table with the returned data.
        """
        if params is None:
            params = {}
        data = Table()
        page = 1
        pages = None

        while pages is None or page <= pages:
            req_params = {**params, "page": page}
            response_data = self.client.get_request(url, req_params)
            pages = response_data["pages"]
            data.concat(Table(response_data[data_key]))

            if large_request:
                data.materialize()

            page += 1

        return data

    def get_folders(self) -> Table:
        """
        Get all folders on your account and their subfolders.

        `Returns:`
            Table Class
                A Table with the folders data.
        """
        response_data = self.client.get_request("folder")
        logger.debug(response_data)

        # The API returns folders in a tree structure that doesn't fit well
        # into a Table. Fortunately, subfolders can't themselves have subfolders.
        subfolders = []
        for f in response_data["folders"]:
            f_subfolders = f.get("subfolders")
            if f_subfolders is not None:
                subfolders += f_subfolders

        tbl = Table(response_data["folders"] + subfolders)
        logger.debug(f"Found {tbl.num_rows} folders.")

        if tbl.num_rows == 0:
            return Table()

        tbl.convert_column("id", int)
        tbl.convert_column("parent", lambda p: None if p == "0" else int(p))
        tbl.remove_column("subfolders")
        return tbl

    def get_forms(self, form_name: Optional[str] = None, folder_id: Optional[int] = None) -> Table:
        """
        Get all forms on your account.

        `Args:`
            form_name: string, optional
                Search by form name.
            folder_id: int, optional
                Return forms in the specified folder.

        `Returns:`
            Table Class
                A table with the forms data.
        """
        params = {}
        if form_name:
            params["search"] = form_name
        if folder_id:
            params["folder"] = folder_id
        response_data = self.client.get_request("form", params)
        logger.debug(response_data)
        return Table(response_data["forms"])

    def get_submission(self, id: int) -> dict:
        """
        Get the details of the specified submission.

        `Args:`
            id: int
                ID for the submission to retrieve.

        `Returns:`
            Dictionary
                Submission data.
        """
        response_data = self.client.get_request(f"submission/{id}")
        logger.debug(response_data)
        return response_data

    def get_form_submissions(self, form_id: int, **query_params) -> Table:
        """
        Get all submissions for the specified form.

        Note this only returns the meta-data about the submissions, not the answer data,
        by default. To get the responses pass `data=True` as a query param.

        For more useful options, such as how to filter the responses by date,
        check the Formstack documentation.

        `Args:`
            form_id: int
                The form ID for the form of the submissions.
            query_params: kwargs
                Query arguments to pass to the form/submissions endpoint.

        `Returns:`
            Table Class
                A Table with the submission data for the form.
        """
        tbl = self._get_paginated_request(
            f"form/{form_id}/submission", "submissions", query_params, True
        )
        logger.debug(tbl)
        return tbl

    def get_form_fields(self, form_id: int) -> Table:
        """
        Get all fields for the specified form.

        `Args:`
            form_id: int
                The form ID for the form of the submissions.

        `Returns:`
            Table Class
                A Table with the fields on the form.
        """
        response_data = self.client.get_request(f"form/{form_id}/field")
        logger.debug(response_data)
        tbl = Table(response_data)
        return tbl



import os
import unittest
from unittest import mock

import pytest
import requests_mock

from parsons import Airmeet, Table

ENV_PARAMETERS = {
    "AIRMEET_URI": "https://env_api_endpoint",
    "AIRMEET_ACCESS_KEY": "env_access_key",
    "AIRMEET_SECRET_KEY": "env_secret_key",
}


class TestAirmeet(unittest.TestCase):
    @requests_mock.Mocker()
    def setUp(self, m):
        m.post("https://api-gateway.airmeet.com/prod/auth", json={"token": "test_token"})
        self.airmeet = Airmeet(airmeet_access_key="fake_key", airmeet_secret_key="fake_secret")
        self.airmeet.client = mock.MagicMock()

    def tearDown(self):
        pass

    @requests_mock.Mocker()
    @mock.patch.dict(os.environ, ENV_PARAMETERS)
    def test_from_environ(self, m):
        m.post("https://env_api_endpoint/auth", json={"token": "test_token"})
        airmeet = Airmeet()
        self.assertEqual(airmeet.uri, "https://env_api_endpoint")
        self.assertEqual(airmeet.airmeet_client_key, "env_access_key")
        self.assertEqual(airmeet.airmeet_client_secret, "env_secret_key")
        self.assertEqual(airmeet.token, "test_token")

    def test_get_all_pages_single_page(self):
        # Simulate API response for a single page without further cursors.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": [{"id": "1", "name": "Item 1"}],
                "cursors": {"pageCount": 1, "after": None},
            }
        )

        url = "airmeet/some_endpoint"
        result = self.airmeet._get_all_pages(url)

        self.airmeet.client.get_request.assert_called_once_with(url=url, params={"size": 50})
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "Table should contain exactly one record"

    def test_get_all_pages_multiple_pages(self):
        # Simulate API responses for multiple pages.
        responses = [
            {
                "data": [{"id": "1", "name": "Item 1"}],
                "cursors": {"pageCount": 2, "after": "abc123"},
            },
            {
                "data": [{"id": "2", "name": "Item 2"}],
                "cursors": {"pageCount": 2, "after": None},
            },  # Last page
        ]
        self.airmeet.client.get_request = mock.MagicMock(
            side_effect=lambda *args, **kwargs: responses.pop(0)
        )

        url = "airmeet/some_endpoint"
        result = self.airmeet._get_all_pages(url)

        calls = [
            mock.call(url=url, params={"size": 50, "after": "abc123"}),
            mock.call(url=url, params={"size": 50, "after": "abc123"}),
        ]
        self.airmeet.client.get_request.assert_has_calls(calls, any_order=True)
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 2, "Table should contain records from both pages"

    def test_list_airmeets(self):
        # Test get the list of Airmeets.
        self.airmeet.client = mock.MagicMock()

        result = self.airmeet.list_airmeets()

        self.airmeet.client.get_request.assert_called_with(
            url="airmeets",
            params={"size": 500},
        )
        assert isinstance(result, Table), "The result should be a Table"

    def test_fetch_airmeet_participants_single_page(self):
        # Simulate API response for a single page of participants. This
        # particular API doesn't use cursors like the other ones that can have
        # multiple pages, which is why this is a separate test.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "participants": [{"user_id": "abc123", "name": "Test User 1"}],
                "userCount": 1,
                "totalUserCount": 1,
            }
        )

        result = self.airmeet.fetch_airmeet_participants("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/participants",
            params={
                "pageNumber": 1,
                "resultSize": 1000,
                "sortingKey": "registrationDate",
                "sortingDirection": "DESC",
            },
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "Table should contain exactly one record"

    def test_fetch_airmeet_participants_multiple_pages(self):
        # Simulate API responses for multiple pages of participants. This
        # particular API doesn't use cursors like the other ones that can have
        # multiple pages, which is why this is a separate test.

        # The connector requests 1000 at a time, so we return a totalUserCount
        # of 2000 here to make it request a second page.
        responses = [
            {
                "participants": [{"user_id": "abc123", "name": "Test User 1"}],
                "userCount": 1,
                "totalUserCount": 2000,
            },
            {
                "participants": [{"user_id": "def456", "name": "Test User 1"}],
                "userCount": 1,
                "totalUserCount": 2000,
            },  # Last page
        ]
        self.airmeet.client.get_request = mock.MagicMock(
            side_effect=lambda *args, **kwargs: responses.pop(0)
        )

        result = self.airmeet.fetch_airmeet_participants("test_airmeet_id")

        calls = [
            mock.call(
                url="airmeet/test_airmeet_id/participants",
                params={
                    "pageNumber": 1,
                    "resultSize": 1000,
                    "sortingKey": "registrationDate",
                    "sortingDirection": "DESC",
                },
            ),
            mock.call(
                url="airmeet/test_airmeet_id/participants",
                params={
                    "pageNumber": 2,
                    "resultSize": 1000,
                    "sortingKey": "registrationDate",
                    "sortingDirection": "DESC",
                },
            ),
        ]
        self.airmeet.client.get_request.assert_has_calls(calls, any_order=True)
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 2, "Table should contain records from both pages"

    def test_fetch_airmeet_sessions(self):
        # Test get the list of sessions for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "name": "Test Event",
                "sessions": [
                    {"sessionid": "test_session_id_1", "name": "Test Session 1"},
                    {"sessionid": "test_session_id_2", "name": "Test Session 2"},
                ],
            }
        )

        result = self.airmeet.fetch_airmeet_sessions("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(url="airmeet/test_airmeet_id/info")
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 2, "Table should contain both records"

    def test_fetch_airmeet_info(self):
        # Test get the Airmeet info.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "name": "Test Event",
                "sessions": [
                    {"sessionid": "test_session_id_1", "name": "Test Session 1"},
                    {"sessionid": "test_session_id_2", "name": "Test Session 2"},
                ],
                "session_hosts": [{"id": "abc123", "name": "Test Host 1"}],
            }
        )

        result = self.airmeet.fetch_airmeet_info("test_airmeet_id", lists_to_tables=True)

        self.airmeet.client.get_request.assert_called_once_with(url="airmeet/test_airmeet_id/info")
        assert isinstance(result, dict), "The result should be a Table"
        assert isinstance(result["sessions"], Table), "The sessions should be a Table"
        assert isinstance(result["session_hosts"], Table), "The session hosts should be a Table"
        assert len(result["sessions"]) == 2, "Sessions Table should contain exactly two records"
        assert len(result["session_hosts"]) == 1, (
            "Session hosts Table should contain exactly one record"
        )

    def test_fetch_airmeet_custom_registration_fields(self):
        # Test get the custom registration fields for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "customFields": [
                    {"fieldId": "test_field_id_1", "label": "Test Label 1"},
                    {"fieldId": "test_field_id_2", "label": "Test Label 2"},
                ],
            }
        )

        result = self.airmeet.fetch_airmeet_custom_registration_fields("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/custom-fields"
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 2, "Table should contain exactly two records"

    def test_fetch_event_attendance(self):
        # Test get the attendees for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": [
                    {
                        "name": "Test User 1",
                        "user_id": "abc123",
                    }
                ],
            }
        )

        result = self.airmeet.fetch_event_attendance("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/attendees",
            params={"size": 50},
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "The result should contain exactly one record"

    def test_fetch_session_attendance(self):
        # Test get the attendees for a session.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": [
                    {
                        "name": "Test User 1",
                        "user_id": "abc123",
                    }
                ],
            }
        )

        result = self.airmeet.fetch_session_attendance("test_session_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="session/test_session_id/attendees",
            params={"size": 50},
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "The result should contain exactly one record"

    def test_fetch_session_attendance_exception_202(self):
        # Test that an asynchronous API raises an exception if it returns
        # a statusCode == 202.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "statusCode": 202,
                "statusMessage": "Preparing your results. Try after 5 minutes"
                "to get the updated results",
            }
        )

        with pytest.raises(Exception):  # noqa: B017
            self.airmeet.fetch_session_attendance("test_session_id")

    def test_fetch_session_attendance_exception_400(self):
        # Test that the sessions attendees API raises an exception if it
        # returns a statusCode == 400.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": {},
                "statusCode": 400,
                "statusMessage": "Session status is not valid",
            }
        )

        with pytest.raises(Exception):  # noqa: B017
            self.airmeet.fetch_session_attendance("test_session_id")

    def test_fetch_airmeet_booths(self):
        # Test get the booths for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "booths": [
                    {
                        "uid": "test_booth_uid_1",
                        "name": "Test Booth 1",
                    }
                ],
            }
        )

        result = self.airmeet.fetch_airmeet_booths("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/booths"
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "The result should contain exactly one record"

    def test_fetch_booth_attendance(self):
        # Test get the attendees for a booth.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": [
                    {
                        "name": "Test User 1",
                        "user_id": "abc123",
                    }
                ],
            }
        )

        result = self.airmeet.fetch_booth_attendance(
            "test_airmeet_id",
            booth_id="test_booth_id",
        )

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/booth/test_booth_id/booth-attendance",
            params={"size": 50},
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "The result should contain exactly one record"

    def test_fetch_poll_responses(self):
        # Test get the poll responses for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": [
                    {
                        "name": "Test User 1",
                        "polls": [
                            {"question": "Poll Question 1", "answer": "Poll Answer 1"},
                            {"question": "Poll Question 2", "answer": "Poll Answer 2"},
                        ],
                    }
                ],
            }
        )

        result = self.airmeet.fetch_poll_responses("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/polls",
            params={"size": 50},
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result[0]["polls"]) == 2, "The record should contain exactly two poll responses"

    def test_fetch_questions_asked(self):
        # Test get the questions asked for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": [
                    {
                        "name": "Test User 1",
                        "questions": [
                            {"question": "Question 1", "session_id": "session_id_1"},
                            {"question": "Question 2", "session_id": "session_id_2"},
                        ],
                    }
                ],
            }
        )

        result = self.airmeet.fetch_questions_asked("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/questions"
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result[0]["questions"]) == 2, "The record should contain exactly two questions"

    def test_fetch_event_tracks(self):
        # Test get the tracks for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "tracks": [
                    {
                        "uid": "test_track_uid_1",
                        "name": "Test Track 1",
                        "sessions": [
                            "session_id_1",
                            "session_id_2",
                        ],
                    }
                ],
            }
        )

        result = self.airmeet.fetch_event_tracks("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/tracks"
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result[0]["sessions"]) == 2, "The record should contain exactly two session ids"

    def test_fetch_registration_utms(self):
        # Test get the registration UTMs for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": [
                    {
                        "airmeetId": "test_airmeet_id",
                        "email": "test@example.com",
                        "id": 1,
                        "utms": {
                            "utm_campaign": "test_utm_campaign",
                            "utm_medium": None,
                            "utm_source": "test_utm_source",
                        },
                    }
                ],
            }
        )

        result = self.airmeet.fetch_registration_utms("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/utms",
            params={"size": 50},
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "The result should contain exactly one record"
        assert len(result[0]["utms"]) == 3, "The record should contain exactly three UTMs"

    def test_download_session_recordings(self):
        # Test get the session recordings for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "recordings": [
                    {
                        "session_id": "test_session_id_1",
                        "session_name": "Test Session 1",
                        "download_link": "https://example.com/test_download_link",
                    }
                ],
            }
        )

        result = self.airmeet.download_session_recordings(
            "test_airmeet_id",
            session_id="test_session_id",
        )

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/session-recordings",
            sessionIds="test_session_id",
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "The result should contain exactly one record"

    def test_fetch_event_replay_attendance(self):
        # Test get the replay attendees for an Airmeet.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": [
                    {
                        "id": 1,
                        "name": "Test User 1",
                        "session_id": "test_session_id",
                    }
                ],
            }
        )

        result = self.airmeet.fetch_event_replay_attendance("test_airmeet_id")

        self.airmeet.client.get_request.assert_called_once_with(
            url="airmeet/test_airmeet_id/event-replay-attendees",
            params={"size": 50},
        )
        assert isinstance(result, Table), "The result should be a Table"
        assert len(result) == 1, "The result should contain exactly one record"

    def test_fetch_event_replay_attendance_exception_400(self):
        # Test that the replay attendees API raises an exception if it returns
        # a statusCode == 400.
        self.airmeet.client.get_request = mock.MagicMock(
            return_value={
                "data": {},
                "statusCode": 400,
                "statusMessage": "Airmeet status is not valid",
            }
        )

        with pytest.raises(Exception):  # noqa: B017
            self.airmeet.fetch_event_replay_attendance("test_airmeet_id")



import unittest

import requests_mock

from parsons import Freshdesk
from test.test_freshdesk import expected_json

DOMAIN = "myorg"
API_KEY = "mykey"


class TestFreshdesk(unittest.TestCase):
    def setUp(self):
        self.fd = Freshdesk(DOMAIN, API_KEY)

    @requests_mock.Mocker()
    def test_get_agents(self, m):
        # Test that agents are returned correctly.
        m.get(self.fd.uri + "agents", json=expected_json.test_agent)
        self.fd.get_agents()

    @requests_mock.Mocker()
    def test_get_tickets(self, m):
        # Test that tickets are returned correctly.
        m.get(self.fd.uri + "tickets", json=expected_json.test_ticket)
        self.fd.get_tickets()

    @requests_mock.Mocker()
    def test_get_companies(self, m):
        # Test that tickets are returned correctly.
        m.get(self.fd.uri + "companies", json=expected_json.test_company)
        self.fd.get_companies()

    @requests_mock.Mocker()
    def test_get_contacts(self, m):
        # Test that tickets are returned correctly.
        m.get(self.fd.uri + "contacts", json=expected_json.test_contact)
        self.fd.get_contacts()

    @requests_mock.Mocker()
    def test_create_ticket(self, m):
        # Import the expected JSON
        from .expected_json import test_create_ticket_response

        # Setup mock response
        m.post(self.fd.uri + "tickets", json=test_create_ticket_response)

        # Call the function
        response = self.fd.create_ticket(
            subject="Support Needed...",
            description="Details about the issue...",
            email="tom@outerspace.com",
            priority=1,
            status=2,
            cc_emails=["ram@freshdesk.com", "diana@freshdesk.com"],
        )

        # Assertions to check if the function behaves as expected
        self.assertEqual(response, test_create_ticket_response)

    @requests_mock.Mocker()
    def test_create_ticket_with_custom_fields(self, m):
        # Import the expected JSON
        from .expected_json import test_create_ticket_with_custom_fields_response

        # Setup mock response
        m.post(self.fd.uri + "tickets", json=test_create_ticket_with_custom_fields_response)

        # Call the function
        response = self.fd.create_ticket(
            subject="Support Needed...",
            description="Details about the issue...",
            email="tom@outerspace.com",
            priority=1,
            status=2,
            cc_emails=["ram@freshdesk.com", "diana@freshdesk.com"],
            custom_fields={"category": "Primary"},
        )

        # Assertions to check if the function behaves as expected
        self.assertEqual(response, test_create_ticket_with_custom_fields_response)



"""NGPVAN Signups Endpoints"""

import logging

from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class Signups(object):
    def __init__(self, van_connection):
        self.connection = van_connection

    def get_signups_statuses(self, event_id=None, event_type_id=None):
        """
        Get a list of valid signup statuses for a given event type
        or event. You must pass one of ``event_id`` or ``event_type_id``
        but not both.

        `Args:`
            event_id: int
                A valid event id.
            event_type_id: int
                A valid event type id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if event_id is None and event_type_id is None:
            raise ValueError("One of event_id or event_type_id must be populated")

        if event_id is not None and event_type_id is not None:
            raise ValueError("Event Id and Event Type ID may not BOTH be populated")

        if event_id:
            params = {"eventId": event_id}
        if event_type_id:
            params = {"eventTypeId": event_type_id}

        tbl = Table(self.connection.get_request("signups/statuses", params=params))
        logger.info(f"Found {tbl.num_rows} signups.")
        return tbl

    def get_person_signups(self, vanid):
        """
        Get the signup history of a person.

        `Args:`
            vanid: int
                A valid vanid associated with a person.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("signups", params={"vanID": vanid}))
        logger.info(f"Found {tbl.num_rows} signups for {vanid}.")
        return self._unpack_signups(tbl)

    def get_event_signups(self, event_id):
        """
        Get the signup history of an event.

        `Args:`
            event_id: int
                A valid event_id associated with an event
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("signups", params={"eventId": event_id}))
        logger.info(f"Found {tbl.num_rows} signups for event {event_id}.")
        return self._unpack_signups(tbl)

    def get_signup(self, event_signup_id):
        """
        Get a single signup object.

        `Args:`
            event_signup_id: int
                A valid event_signup_id associated with a signup.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        r = self.connection.get_request(f"signups/{event_signup_id}")
        logger.info(f"Found sign up {event_signup_id}.")
        return r

    def create_signup(self, vanid, event_id, shift_id, role_id, status_id, location_id):
        """
        Create a new signup for an event.

        `Args:`
            vanid: int
                A valid vanid of the person to signup for the event.
            event_id: int
                A valid event_id to associate the person with the event
            shift_id:
                A shift_id, associated with the event to assign the person
            role_id:
                A role_id, associated with the event to assign the person
            status_id:
                A status_id of the person
            location_id:
                A location_id for the event
        `Returns:`
            Int
                The event signup id
        """

        signup = {
            "person": {"vanId": vanid},
            "event": {"eventId": event_id},
            "shift": {"eventShiftId": shift_id},
            "role": {"roleId": role_id},
            "status": {"statusId": status_id},
            "location": {"locationId": location_id},
        }

        r = self.connection.post_request("signups", json=signup)
        logger.info(f"Signup {r} created.")
        return r

    def update_signup(
        self,
        event_signup_id,
        shift_id=None,
        role_id=None,
        status_id=None,
        location_id=None,
    ):
        """
        Update a signup object. All of the kwargs will update the values associated
        with them.

        `Args:`
            event_signup_id: int
                A valid event signup id
            shift_id: int
                The shift_id to update
            role_id: int
                The role_id to update
            status_id: int
                The status_id to update
            location_id: int
                The location_id to update
        `Returns:`
            ``None``
        """

        #  Get the signup object
        signup = self.connection.get_request(f"signups/{event_signup_id}")

        # Update the signup object
        if shift_id:
            signup["shift"] = {"eventShiftId": shift_id}
        if role_id:
            signup["role"] = {"roleId": role_id}
        if status_id:
            signup["status"] = {"statusId": status_id}
        if location_id:
            signup["location"] = {"locationId": location_id}

        return self.connection.put_request(f"signups/{event_signup_id}", json=signup)

    def delete_signup(self, event_signup_id):
        """
        Delete a signup object

        `Args:`
            event_signup_id: int
                A valid event signup id
        `Returns:`
            ``None``
        """

        r = self.connection.delete_request(f"signups/{event_signup_id}")
        logger.info(f"Signup {event_signup_id} deleted.")
        return r

    def _unpack_signups(self, table):
        # Unpack all of the nested jsons
        table.unpack_dict("person", prepend=False)
        table.unpack_dict("status")
        table.unpack_dict("event")
        table.unpack_dict("shift")
        table.unpack_dict("role")
        table.unpack_dict("location")

        return table



import os
import unittest.mock

import petl
import requests_mock

from parsons import VAN, Table
from test.utils import assert_matching_tables, validate_list

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestTargets(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    def tearDown(self):
        pass

    @requests_mock.Mocker()
    def test_get_targets(self, m):
        # Create response
        json = {
            "count": 2,
            "items": [
                {
                    "targetId": 12827,
                    "type": "TEST CODE",
                    "name": "TEST CODE",
                    "description": None,
                    "points": 20,
                    "areSubgroupsSticky": False,
                    "status": "Active",
                    "subgroups": None,
                    "markedSubgroup": None,
                }
            ],
            "nextPageLink": None,
        }

        m.get(self.van.connection.uri + "targets", json=json)

        # Expected Structure
        expected = [
            "targetId",
            "type",
            "name",
            "description",
            "points",
            "areSubgroupsSticky",
            "status",
            "subgroups",
            "markedSubgroup",
        ]

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.van.get_targets()))

        # To Do: Test what happens when it doesn't find any targets

    @requests_mock.Mocker()
    def test_get_target(self, m):
        # Create response
        json = {
            "targetId": 15723,
            "name": "Mail_VR_Chase",
            "type": "Dynamic",
            "description": None,
            "points": 15,
            "areSubgroupsSticky": False,
            "status": "Active",
            "subgroups": [
                {
                    "targetId": 12827,
                    "fullName": "April_VR_Chase Calls",
                    "name": "April_Chase_20",
                    "subgroupId": 46803,
                    "isAssociatedWithBadges": True,
                }
            ],
            "markedSubgroup": None,
        }

        m.get(self.van.connection.uri + "targets/15723", json=json)

        self.assertEqual(json, self.van.get_target(15723))

    @requests_mock.Mocker()
    def test_create_target_export(self, m):
        export_job_id = '{"exportJobId": "455961790"}'
        target_id = 12827

        m.post(
            self.van.connection.uri + "targetExportJobs",
            json=export_job_id,
            status_code=204,
        )

        # Test that it doesn't throw and error
        r = self.van.create_target_export(target_id, webhook_url=None)

        self.assertEqual(r, export_job_id)

    @unittest.mock.patch.object(petl, "fromcsv", autospec=True)
    @requests_mock.Mocker()
    def test_get_target_export(self, fromcsv, m):
        export_job_id = 455961790
        json = {
            "targetId": 12827,
            "file": {
                "downloadUrl": (
                    "https://ngpvan.blob.core.windows.net/"
                    "target-export-files/TargetExport_455961790.csv"
                ),
                "dateExpired": "null",
                "recordCount": 1016883,
            },
            "webhookUrl": "null",
            "exportJobId": 455961790,
            "jobStatus": "Complete",
        }

        download_url = (
            "https://ngpvan.blob.core.windows.net/target-export-files/TargetExport_455961790.csv"
        )
        fromcsv.return_value = petl.fromcolumns(
            [
                ["12827", "12827"],
                ["Volunteer Recruitment Tiers", "Volunteer Recruitment Tiers"],
                ["1111", "1111"],
                ["Tier", "Tier"],
                ["109957749", "109957754"],
            ],
            [
                "TargetID",
                "TargetName",
                "TargetSubgroupID",
                "TargetSubgroupName",
                "VanID",
            ],
        )

        m.post(
            self.van.connection.uri + "targetExportJobs",
            json=export_job_id,
            status_code=204,
        )
        m.get(self.van.connection.uri + "targetExportJobs/455961790", json=json)

        expected_result = Table(
            [
                (
                    "TargetID",
                    "TargetName",
                    "TargetSubgroupID",
                    "TargetSubgroupName",
                    "VanID",
                ),
                ("12827", "Volunteer Recruitment Tiers", "1111", "Tier", "109957749"),
                ("12827", "Volunteer Recruitment Tiers", "1111", "Tier", "109957754"),
            ]
        )

        assert_matching_tables(self.van.get_target_export(export_job_id), expected_result)
        self.assertEqual(fromcsv.call_args, unittest.mock.call(download_url, encoding="utf-8-sig"))



import logging

from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class Email(object):
    """
    Instantiate the Email class.

    You can find the docs for the NGP VAN Email API here:
    https://docs.ngpvan.com/reference/email-overview
    """

    def __init__(self, van_connection):
        self.connection = van_connection

    def get_emails(self, ascending: bool = True) -> Table:
        """
        Get emails.

        `Args:`
            ascending : Bool
                sorts results in ascending or descending order
                for the dateModified field. Defaults to True (ascending).

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        if ascending:
            params = {
                "$orderby": "dateModified asc",
            }
        if not ascending:
            params = {
                "$orderby": "dateModified desc",
            }

        tbl = Table(self.connection.get_request("email/messages", params=params))
        logger.debug(f"Found {tbl.num_rows} emails.")
        return tbl

    def get_email(self, email_id: int, expand: bool = True) -> Table:
        """
        Get an email.

        Note that it takes some time for the system to aggregate opens and click-throughs,
        so data can be delayed up to 15 minutes.

        `Args:`
            email_id : int
                The email id.
            expand : bool
                Optional; expands the email message to include the email content and
                statistics. Defaults to True.

        `Returns:`
            dict
        """

        params = {
            "$expand": (
                "emailMessageContent, EmailMessageContentDistributions" if expand else None
            ),
        }

        r = self.connection.get_request(f"email/message/{email_id}", params=params)
        logger.debug(f"Found email {email_id}.")
        return r

    def get_email_stats(self, aggregate_ab: bool = True) -> Table:
        """
        Get stats for all emails, aggregating any A/B tests.

        Note: Pending emails will have a dateScheduled of "0001-01-01T00:00:00Z"
        and a subject line of "None". This is a limitation of the NGPVAN API.
        Also note that any information on opens, clicks, etcetera will default to 0.

        `Args:`
            aggregate_ab : bool
                If A/B test results for emails should get aggregated.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        email_list = []

        final_email_list = []

        emails = self.get_emails()

        foreign_message_ids = [email["foreignMessageId"] for email in emails]

        for fmid in foreign_message_ids:
            email = self.get_email(fmid)
            email_list.append(email)

        # Outside and inside emailMessageContentDistributions field
        outer_fields = [
            "name",
            "createdBy",
            "dateCreated",
            "dateModified",
            "dateScheduled",
            "foreignMessageId",
        ]
        inner_fields = [
            "recipientCount",
            "bounceCount",
            "contributionCount",
            "contributionTotal",
            "formSubmissionCount",
            "linksClickedCount",
            "machineOpenCount",
            "openCount",
            "unsubscribeCount",
            # "subject"  # included here for clarity, but has some special logic
        ]
        # If we are aggregating, we have one entry per foreignMessageId (outer loop) and
        # sum over the values inside the inner loop. If we are not, then we need to loop
        # over foreignMessageId and each component of emailMessageContent, so we loop
        # over both and pull out data (with no aggregation) for each.
        if aggregate_ab:
            for email in email_list:  # One row per foreignMessageId
                outer = {field: email[field] for field in outer_fields}
                inner = dict.fromkeys(inner_fields, 0)
                for i in email["emailMessageContent"]:
                    # Pending emails don't have emailMessageContentDistributions, just have defaults
                    if not i["emailMessageContentDistributions"]:
                        logger.info(
                            f"No emailMessageContentDistributions for email {i['name']}, defaulting values to 0"
                        )
                    else:
                        try:
                            for field in inner_fields:  # Aggregation of all inner values
                                inner[field] += i["emailMessageContentDistributions"][field]
                            # Just replacing subject to get the last one
                            inner["subject"] = i["subject"]
                        except KeyError as e:
                            logger.info(str(e))
                            pass
                final_email_list.append({**outer, **inner})
        else:
            for email in email_list:
                for i in email["emailMessageContent"]:
                    # One row per foreignMessageId / emailMessageContent entry
                    outer = {field: email[field] for field in outer_fields}
                    inner = dict.fromkeys(inner_fields, 0)
                    if not i["emailMessageContentDistributions"]:
                        logger.info(
                            f"No emailMessageContentDistributions for email {i['name']}, defaulting values to 0"
                        )
                    else:
                        try:
                            for field in inner_fields:
                                inner[field] = i["emailMessageContentDistributions"][field]
                            inner["subject"] = i["subject"]
                        except KeyError as e:
                            logger.info(str(e))
                    final_email_list.append({**outer, **inner})

        return Table(final_email_list)



import os
import unittest

import requests_mock
from requests.exceptions import HTTPError

from parsons import VAN
from test.utils import validate_list

os.environ["VAN_API_KEY"] = "SOME_KEY"


location_json = {
    "locationId": 34,
    "name": "Chicagowide",
    "displayName": "Chicagowide, Chicago, IL ",
    "address": {
        "addressId": None,
        "addressLine1": None,
        "addressLine2": None,
        "addressLine3": None,
        "city": "Chicago",
        "stateOrProvince": "IL",
        "zipOrPostalCode": None,
        "geoLocation": None,
        "countryCode": "US",
        "preview": "Chicago, IL ",
        "type": None,
        "isPreferred": None,
        "streetAddress": None,
        "displayMode": "Standardized",
    },
    "id": 34,
    "notes": None,
    "codes": None,
}

expected_loc = [
    "locationId",
    "name",
    "displayName",
    "id",
    "notes",
    "codes",
    "addressId",
    "addressLine1",
    "addressLine2",
    "addressLine3",
    "city",
    "countryCode",
    "displayMode",
    "isPreferred",
    "preview",
    "stateOrProvince",
    "streetAddress",
    "type",
    "zipOrPostalCode",
]


class TestLocations(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="EveryAction")

    def tearDown(self):
        pass

    @requests_mock.Mocker()
    def test_get_locations(self, m):
        json = {"items": [location_json], "nextPageLink": None, "count": 1}
        m.get(self.van.connection.uri + "locations", json=json)

        self.assertTrue(validate_list(expected_loc, self.van.get_locations()))

    @requests_mock.Mocker()
    def test_get_location(self, m):
        # Valid location id
        m.get(self.van.connection.uri + "locations/34", json=location_json)
        self.assertEqual(location_json, self.van.get_location(34))

    @requests_mock.Mocker()
    def test_delete_location(self, m):
        # Test good location delete
        m.delete(self.van.connection.uri + "locations/1", status_code=200)
        self.van.delete_location(1)

        # Test invalid location delete
        m.delete(self.van.connection.uri + "locations/2", status_code=404)
        self.assertRaises(HTTPError, self.van.delete_location, 2)

    @requests_mock.Mocker()
    def test_create_location(self, m):
        loc_id = 32

        m.post(
            self.van.connection.uri + "locations/findOrCreate",
            json=loc_id,
            status_code=204,
        )

        self.assertEqual(
            self.van.create_location(name="Chicagowide", city="Chicago", state="IL"),
            loc_id,
        )



"""NGPVAN Changed Entities"""

import logging
import time

from parsons.etl.table import Table

logger = logging.getLogger(__name__)

RETRY_RATE = 10


class ChangedEntities(object):
    def __init__(self):
        pass

    def get_changed_entity_resources(self):
        """
        Get changed entity resources available to the API user.

        `Returns:`
            list
        """

        r = self.connection.get_request("changedEntityExportJobs/resources")
        logger.info(f"Found {len(r)} changed entity resources.")
        return r

    def get_changed_entity_resource_fields(self, resource_type):
        """
        Get export fields avaliable for each changed entity resource.

        `Args:`
            resource_type: str
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request(f"changedEntityExportJobs/fields/{resource_type}"))
        logger.info(f"Found {tbl.num_rows} fields for {resource_type}.")
        return tbl

    def get_changed_entities(
        self,
        resource_type,
        date_from,
        date_to=None,
        include_inactive=False,
        requested_fields=None,
        custom_fields=None,
    ):
        """
        Get modified records for VAN from up to 90 days in the past.

        `Args:`
            resource_type: str
                The type of resource to export. Use the :py:meth:`~parsons.ngpvan.changed_entities.ChangedEntities.get_changed_entity_resources`
                to get a list of potential entities.
            date_from: str
                The start date in which to search. Must be less than 90 days in the
                past. Must be``iso8601`` formatted date (``2021-10-11``).
            date_to: str
                The end date to search. Must be less than 90 days in the
                past. Must be``iso8601`` formatted date (``2021-10-11``).
            include_inactive: boolean
                Include inactive records
            requested_fields: list
                A list of optional requested fields to include. These options can be accessed through
                :py:meth:`~parsons.ngpvan.changed_entities.ChangedEntities.get_changed_entity_resource_fields`
                method.
            custom_fields: list
                A list of ids of custom fields to include in the export.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        json = {
            "dateChangedFrom": date_from,
            "dateChangedTo": date_to,
            "resourceType": resource_type,
            "requestedFields": requested_fields,
            "requestedCustomFieldIds": custom_fields,
            "fileSizeKbLimit": 100000,
            "includeInactive": include_inactive,
        }

        r = self.connection.post_request("changedEntityExportJobs", json=json)

        while True:
            status = self._get_changed_entity_job(r["exportJobId"])
            if status["jobStatus"] in ["Pending", "InProcess"]:
                logger.info("Waiting on export file.")
                time.sleep(RETRY_RATE)
            elif status["jobStatus"] == "Complete":
                return Table.from_csv(status["files"][0]["downloadUrl"])
            else:
                raise ValueError(status["message"])

    def _get_changed_entity_job(self, job_id):
        r = self.connection.get_request(f"changedEntityExportJobs/{job_id}")
        return r



import os
from distutils.core import setup
from pathlib import Path

from setuptools import find_packages


def main():
    limited_deps = os.environ.get("PARSONS_LIMITED_DEPENDENCIES", "")
    if limited_deps.strip().upper() in ("1", "YES", "TRUE", "ON"):
        install_requires = [
            "petl",
            "python-dateutil",
            "requests",
            "requests_oauthlib",
            "simplejson",
        ]
        extras_require = {
            "airtable": ["pyairtable"],
            "alchemer": ["surveygizmo"],
            "azure": ["azure-storage-blob"],
            "box": ["boxsdk"],
            "braintree": ["braintree"],
            "catalist": ["paramiko"],
            "civis": ["civis"],
            "dbt-redshift": ["dbt-redshift>=1.5.0"],
            "dbt-bigquery": ["dbt-bigquery>=1.5.0"],
            "dbt-postgres": ["dbt-postgres>=1.5.0"],
            "dbt-snowflake": ["dbt-snowflake>=1.5.0"],
            "facebook": ["joblib", "facebook-business"],
            "geocode": ["censusgeocode", "urllib3==1.26.19"],
            "github": ["PyGitHub"],
            "google": [
                "apiclient",
                "google-api-python-client",
                "google-cloud-bigquery",
                "google-cloud-storage",
                "google-cloud-storage-transfer",
                "gspread",
                "httplib2",
                "oauth2client",
                "validate-email",
            ],
            "mysql": [
                "mysql-connector-python",
                "sqlalchemy >= 1.4.22, != 1.4.33, < 3.0.0",
            ],
            "newmode": ["newmode"],
            "ngpvan": ["suds-py3"],
            "mobilecommons": ["bs4"],
            "postgres": [
                "psycopg2-binary<=2.9.9;python_version<'3.13'",
                "psycopg2-binary>=2.9.10;python_version>='3.13'",
                "sqlalchemy >= 1.4.22, != 1.4.33, < 3.0.0",
            ],
            "redshift": [
                "boto3",
                "psycopg2-binary<=2.9.9;python_version<'3.13'",
                "psycopg2-binary>=2.9.10;python_version>='3.13'",
                "sqlalchemy >= 1.4.22, != 1.4.33, < 3.0.0",
            ],
            "s3": ["boto3"],
            "salesforce": ["simple-salesforce"],
            "scytl": ["defusedxml", "pytz"],
            "sftp": ["paramiko"],
            "slack": ["slackclient<2"],
            "smtp": ["validate-email"],
            "targetsmart": ["xmltodict", "defusedxml"],
            "twilio": ["twilio"],
            "ssh": [
                "sshtunnel",
                "psycopg2-binary<=2.9.9;python_version<'3.13'",
                "psycopg2-binary>=2.9.10;python_version>='3.13'",
                "sqlalchemy >= 1.4.22, != 1.4.33, < 3.0.0",
            ],
        }
        extras_require["all"] = sorted({lib for libs in extras_require.values() for lib in libs})
    else:
        THIS_DIR = os.path.abspath(os.path.dirname(__file__))
        with open(os.path.join(THIS_DIR, "requirements.txt")) as reqs:
            install_requires = reqs.read().strip().split("\n")
        # No op for forward-compatibility
        extras_require = {"all": []}

    this_directory = Path(__file__).parent
    long_description = (this_directory / "README.md").read_text()

    setup(
        name="parsons",
        version="5.1.0",  # ensure this version number is in format of n.n.n, not n or n.n
        author="The Movement Cooperative",
        author_email="info@movementcooperative.org",
        url="https://github.com/move-coop/parsons",
        keywords=["PROGRESSIVE", "API", "ETL"],
        packages=find_packages(),
        install_requires=install_requires,
        extras_require=extras_require,
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
        ],
        python_requires=">=3.9,<3.14",
        long_description=long_description,
        long_description_content_type="text/markdown",
    )


if __name__ == "__main__":
    main()



"""Utilities for working with the Catalist Match API

Install dependencies with `pip install parsons[catalist]`
"""

import base64
import logging
import os
import tempfile
import time
import urllib
from typing import Dict, List, Optional, Union
from zipfile import ZipFile

from parsons.etl import Table
from parsons.sftp import SFTP
from parsons.utilities.oauth_api_connector import OAuth2APIConnector

logger = logging.getLogger(__name__)

# Default byte size to export under the hood via Paramiko
DEFAULT_EXPORT_CHUNK_SIZE = 1024 * 1024 * 50


class CatalistMatch:
    """Connector for working with the Catalist Match API.

    This API allows a trusted third party to submit new files for processing, and/or
    reprocess existing files. It also allows retrieval of processing status. Initial
    setup of template(s) via the M Tool UI will be required.

    The Catalist Match tool requires OAuth2.0 client credentials for the API as well as
    credentials for accessing the Catalist sftp bucket. Each Catalist client is given
    their own bucket alias named after a tree species, used for constructing the
    filepath within the sftp bucket.

    Accessing the Catalist sftp bucket and Match API both require the source IP address
    to be explicitly white-listed by Catalist.

    Example usage:
    ```
    tbl = Table.from_csv(...)
    client = CatalistMatch(...)
    match_result = client.match(tbl)
    ```

    Note that matching can take from 10 minutes up to 6 hours or longer to complete, so
    you may want to think strategically about how to await completion without straining
    your compute resources on idling.

    To separate submitting the job and fetching the result:
    ```
    tbl = Table.from_csv(...)
    client = CatalistMatch(...)
    response = client.upload(tbl)
    match_result = client.await_completion(response["id"])
    ```

    """

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        sftp_username: str,
        sftp_password: str,
        client_audience: Optional[str] = None,
    ) -> None:
        self.client_id = client_id
        self.client_secret = client_secret
        self.connection = OAuth2APIConnector(
            "https://api.catalist.us/mapi/",
            client_id=client_id,
            client_secret=client_secret,
            authorization_kwargs={"audience": client_audience or "catalist_api_m_prod"},
            token_url="https://auth.catalist.us/oauth/token",
            auto_refresh_url="https://auth.catalist.us/oauth/token",
        )
        self.sftp = SFTP("t.catalist.us", sftp_username, sftp_password, timeout=7200)

    def load_table_to_sftp(self, table: Table, input_subfolder: Optional[str] = None) -> str:
        """Load table to Catalist sftp bucket as gzipped CSV for matching.

        If input_subfolder is specific, the file will be uploaded to a subfolder of the
        myUploads directory in the SFTP server.

        `Args:`
             table: Table
                 Parsons Table for matching. "first_name" and "last_name" columns
                 are required. Optional columns for matching: last_name, name_suffix,
                 addr1, addr2, city, state, zip, phone, email, gender_tomatch, dob,
                 dob_year, matchbackid.
             input_subfolder: str
                 Optional. If specified, the file will be uploaded to a subfolder of the
                 myUploads directory in the SFTP server.
        """
        local_path = table.to_csv(temp_file_compression="gzip")
        hashed_name = hash(time.time())
        remote_path_parts = ["myUploads", f"{hashed_name}.csv.gz"]
        if input_subfolder:
            if input_subfolder not in self.sftp.list_directory("/myUploads/"):
                self.sftp.make_directory("/myUploads/" + input_subfolder)
            remote_path_parts.insert(1, input_subfolder)
        remote_path = "/".join(remote_path_parts)

        self.sftp.put_file(local_path, remote_path)

        # Loads to Catalist SFTP bucket are expcted in the client's uploads bucket
        # So we don't need to explicitly include that part of the path
        result = f"file://{remote_path.replace('myUploads/', '')}"
        return result

    def match(
        self,
        table: Table,
        export: bool = False,
        description: Optional[str] = None,
        export_filename_suffix: Optional[str] = None,
        input_subfolder: Optional[str] = None,
        copy_to_sandbox: bool = False,
        static_values: Optional[Dict[str, Union[str, int]]] = None,
        wait: int = 30,
    ) -> Table:
        """Load table to the Catalist Match API, returns matched table.

         This method blocks until the match completes, which can take from 10 minutes to
         6 hours or more depending on concurrent traffic.

        `Args:`
             table: Table
                 Parsons Table for matching. "first_name" and "last_name" columns
                 are required. Optional columns for matching: last_name, name_suffix,
                 addr1, addr2, city, state, zip, phone, email, gender_tomatch, dob,
                 dob_year, matchbackid.
             export: bool
                 Defaults to False
             description: str
                 Optional. Description for the match job.
             export_filename_suffix: str
                 Optional. Adds a suffix to the result filename in the SFTP server.
             input_subfolder: str
                 Optional. Adds a prefix to the filepath of the uploaded input file in
                 the SFTP server.
             copy_to_sandbox: bool
                  Defaults to False.
             static_values: dict
                  Optional. Any included values are mapped to every row of the input table.
             wait: int
                  Seconds to poll, defaults to 30.
        """
        response = self.upload(
            table=table,
            export=export,
            description=description,
            export_filename_suffix=export_filename_suffix,
            input_subfolder=input_subfolder,
            copy_to_sandbox=copy_to_sandbox,
            static_values=static_values,
        )
        result = self.await_completion(response["id"], wait=wait)
        return result

    def upload(
        self,
        table: Table,
        template_id: str = "48827",
        export: bool = False,
        description: Optional[str] = None,
        export_filename_suffix: Optional[str] = None,
        input_subfolder: Optional[str] = None,
        copy_to_sandbox: bool = False,
        static_values: Optional[Dict[str, Union[str, int]]] = None,
    ) -> dict:
        """Load table to the Catalist Match API, returns response with job metadata.

        `Args:`
             table: Table
                 Parsons Table for matching. "first_name" and "last_name" columns
                 are required. Optional columns for matching: last_name, name_suffix,
                 addr1, addr2, city, state, zip, phone, email, gender_tomatch, dob,
                 dob_year, matchbackid.
             template_id: str
                 Defaults to 48827, currently the only available template for working
                 with the Match API.
             export: bool
                 Defaults to False
             description: str
                 Optional. Description for the match job.
             export_filename_suffix: str
                 Optional. Adds a suffix to the result filename in the SFTP server.
             input_subfolder: str
                 Optional. Adds a prefix to the filepath of the uploaded input file in
                 the SFTP server.
             copy_to_sandbox: bool
                  Defaults to False.
             static_values: dict
                  Optional. Any included values are mapped to every row of the input table.
        """

        self.validate_table(table, template_id)

        # upload table to s3 temp location
        sftp_file_path = self.load_table_to_sftp(table, input_subfolder)
        sftp_file_path_encoded = base64.b64encode(sftp_file_path.encode("ascii")).decode("ascii")

        if export:
            action = "export%2Cpublish"
        else:
            action = "publish"

        # Create endpoint using options
        endpoint_params = [
            "upload",
            "template",
            template_id,
            "action",
            action,
            "url",
            sftp_file_path_encoded,
        ]

        if description:
            endpoint_params.extend(["description", description])

        endpoint = "/".join(endpoint_params)

        # Assemble query parameters
        query_params: Dict[str, Union[str, int]] = {"token": self.connection.token["access_token"]}
        if copy_to_sandbox:
            query_params["copyToSandbox"] = "true"
        if static_values:
            query_params.update(static_values)
        if export_filename_suffix:
            query_params["subClientName"] = export_filename_suffix

        logger.debug(f"Executing request to endpoint {self.connection.uri + endpoint}")

        endpoint = endpoint + "?" + urllib.parse.urlencode(query_params)

        response = self.connection.get_request(endpoint)

        result = response[0]

        return result

    def action(
        self,
        file_ids: Union[str, List[str]],
        match: bool = False,
        export: bool = False,
        export_filename_suffix: Optional[str] = None,
        copy_to_sandbox: bool = False,
    ) -> List[dict]:
        """Perform actions on existing files.

        All files must be in Finished status (if the action requested is publish), and
        must mapped against the same template. The request will return as soon as the
        action has been queued.

        `Args:`
             file_ids: str or List[str]
                 one or more file_ids (found in the `id` key of responses from the
                 upload() or status() methods)
             match: bool
                 Optional. Defaults to False. If True, will initiate matching.
             export: bool
                 Optional. Defaults to False. If True, will initiate export.
             export_filename_suffix: str
                 Optional. If included, adds a suffix to the filepath of the exported
                 file in the SFTP server.
             copy_to_sandbox: bool
                  Defaults to False.

        """
        actions = ["publish"]
        if match:
            actions.append("match")
        if export:
            actions.append("export")
        action = urllib.parse.quote(",".join(actions))

        endpoint_params = ["action", action]

        if isinstance(file_ids, list):
            encoded_files = urllib.parse.quote(",".join(file_ids))
        else:
            encoded_files = file_ids

        endpoint_params.extend(["file", encoded_files])

        endpoint = "/".join(endpoint_params)

        logger.debug(f"Executing request to endpoint {self.connection.uri + endpoint}")

        query_params = {"token": self.connection.token["access_token"]}
        if copy_to_sandbox:
            query_params["copyToSandbox"] = "true"
        if export_filename_suffix:
            query_params["subClientName"] = export_filename_suffix

        endpoint = endpoint + "?" + urllib.parse.urlencode(query_params)

        result = self.connection.get_request(endpoint)

        return result

    def status(self, id: str) -> dict:
        """Check status of a match job."""
        endpoint = "/".join(["status", "id", id])
        query_params = {"token": self.connection.token["access_token"]}
        result = self.connection.get_request(endpoint, params=query_params)
        return result

    def await_completion(
        self,
        id: str,
        wait: int = 30,
    ) -> Table:
        """
        Await completion of a match job. Return matches when ready.

        This method will poll the status of a match job on a timer until the job is
        complete. By default, polls once every 30 seconds.

        Note that match job completion can take from 10 minutes up to 6 hours or more
        depending on concurrent traffic. Consider your strategy for polling for
        completion.
        """

        while True:
            response = self.status(id)
            status = response["process"]["processState"]
            if status in ("Finished", "Error", "Stopped", "Exception"):
                logger.info(f"Job {id} is complete with status {status}.")
                break

            logger.info(f"Job {id} has status {status}, awaiting completion.")
            time.sleep(wait)

        result = self.load_matches(id=id)
        return result

    def load_matches(self, id: str) -> Table:
        """Take a completed job ID, download and open the match file as a Table.

        Result will be a Table with all the original columns along with columns 'DWID',
        'CONFIDENCE', 'ZIP9', and 'STATE'. The original column headers will be prepended
        with 'COL#-'."""
        # Validate that the job is complete
        response = self.status(str(id))
        status = response["process"]["processState"]

        if status == "Finished":
            logger.info(f"Validated that job {id} completed successfully.")
        else:
            err_msg = "Failed to successfully run match job. "
            if status == "Error":
                err_msg += "Internal error. "
            elif status == "Stopped":
                err_msg += "Probably stopped by Catalist staff. Will be rerun. "
            elif status == "Exception":
                err_msg += (
                    "Error with data. Catalist will have been notified and "
                    "will contact you or rerun the file. "
                )
            else:
                "Unknown or unexpected final status."
            err_msg += f"[job={id}, final_status={status}]"
            raise RuntimeError(err_msg)

        remote_filepaths = self.sftp.list_directory("/myDownloads/")
        remote_filename = [filename for filename in remote_filepaths if id in filename][0]
        remote_filepath = "/myDownloads/" + remote_filename
        temp_file_zip = self.sftp.get_file(
            remote_path=remote_filepath, export_chunk_size=DEFAULT_EXPORT_CHUNK_SIZE
        )
        temp_dir = tempfile.mkdtemp()

        with ZipFile(temp_file_zip) as zf:
            zf.extractall(path=temp_dir)

        filepath = os.listdir(temp_dir)[0]

        result = Table.from_csv(os.path.join(temp_dir, filepath), delimiter="\t")
        return result

    def validate_table(self, table: Table, template_id: str = "48827") -> None:
        """Validate table structure and contents."""
        if not template_id == "48827":
            logger.warn(f"No validator implemented for template {template_id}.")
            return

        expected_table_columns = [
            "first_name",
            "middle_name",
            "last_name",
            "name_suffix",
            "addr1",
            "addr2",
            "city",
            "state",
            "zip",
            "phone",
            "email",
            "gender_tomatch",
            "dob",
            "dob_year",
            "matchbackid",
        ]

        required_columns: List[str] = ["first_name", "last_name"]
        actual_table_columns = table.columns

        unexpected_columns = [
            col for col in actual_table_columns if col not in expected_table_columns
        ]
        missing_required_columns = [
            col for col in required_columns if col not in actual_table_columns
        ]

        errors = {}
        if unexpected_columns:
            errors["unexpected_columns"] = unexpected_columns
        if missing_required_columns:
            errors["missing_required_columns"] = missing_required_columns

        if errors:
            raise ValueError("Input table does not have the right structure. %s", errors)
        else:
            logger.info("Table structure validated.")



from parsons.etl.table import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

AIRMEET_DEFAULT_URI = "https://api-gateway.airmeet.com/prod/"


class Airmeet(object):
    """
    Instantiate class.

       `Args:`
            airmeet_uri: string
                The URI of the Airmeet API endpoint. Not required. The default
                is https://api-gateway.airmeet.com/prod/. You can set an
                ``AIRMEET_URI`` env variable or use this parameter when
                instantiating the class.
            airmeet_access_key: string
                The Airmeet API access key.
            airmeet_secret_key: string
                The Airmeet API secret key.

        For instructions on how to generate an access key and secret key set,
        see `Airmeet's Event Details API documentation
        <https://help.airmeet.com/support/solutions/articles/82000909768-1-event-details-airmeet-public-api>`_.
    """

    def __init__(self, airmeet_uri=None, airmeet_access_key=None, airmeet_secret_key=None):
        """
        Authenticate with the Airmeet API and update the connection headers
        with the access token.

        `Args:`
            airmeet_uri: string
                The Airmeet API endpoint.
            airmeet_access_key: string
                The Airmeet API access key.
            airmeet_secret_key: string
                The Airmeet API secret key.
        `Returns:`
            ``None``
        """
        self.uri = check_env.check("AIRMEET_URI", airmeet_uri, optional=True) or AIRMEET_DEFAULT_URI
        self.client = APIConnector(self.uri)
        self.airmeet_client_key = check_env.check("AIRMEET_ACCESS_KEY", airmeet_access_key)
        self.airmeet_client_secret = check_env.check("AIRMEET_SECRET_KEY", airmeet_secret_key)
        self.client.headers = {
            "X-Airmeet-Access-Key": self.airmeet_client_key,
            "X-Airmeet-Secret-Key": self.airmeet_client_secret,
        }
        response = self.client.post_request(url="auth", success_codes=[200])
        self.token = response["token"]

        # API calls expect the token in the header.
        self.client.headers = {
            "Content-Type": "application/json",
            "X-Airmeet-Access-Token": self.token,
        }

    def _get_all_pages(self, url, page_size=50, **kwargs) -> Table:
        """
        Get all the results from an Airmeet API url, handling pagination based
        on the returned pageCount.

        `Args:`
            page_size: 50
                The number of items to get per page. The max allowed varies by
                API call. For details, see `Airmeet's Event Details API
                documentation
                <https://help.airmeet.com/support/solutions/articles/82000909768-1-event-details-airmeet-public-api>`_.
            **kwargs:
                Additional parameters to include in the request.
        `Returns:`
            ``None``
        """
        results = []
        cursor_after = ""  # For getting the next set of results
        kwargs["size"] = page_size

        # Initial API call to get the first page of data
        response = self.client.get_request(url=url, params=kwargs)

        # Some APIs are asynchronous and will return a 202 if the request
        # should be tried again after five minutes, because the results
        # set needs to be built.
        if "statusCode" in response and response["statusCode"] != 200:
            raise Exception(response)
        else:
            results.extend(response["data"])

            if "cursors" in response and response["cursors"]["pageCount"] > 1:
                cursor_after = response["cursors"]["after"]

                # Fetch subsequent pages if needed
                for _ in range(2, response["cursors"]["pageCount"] + 1):
                    kwargs["after"] = cursor_after
                    response = self.client.get_request(url=url, params=kwargs)
                    results.extend(response["data"])
                    cursor_after = response["cursors"]["after"]

        return Table(results)

    def list_airmeets(self) -> Table:
        """
        Get the list of Airmeets. The API excludes any Airmeets that are
        Archived (Deleted).

        `Returns:`
            Parsons.Table
                List of Airmeets
        """
        return self._get_all_pages(url="airmeets", page_size=500)

    def fetch_airmeet_participants(
        self, airmeet_id, sorting_key="registrationDate", sorting_direction="DESC"
    ) -> Table:
        """
        Get all participants (registrations) for a specific Airmeet, handling
        pagination based on the returned totalUserCount. This API doesn't use
        cursors for paging, so we can't use _get_all_pages() here.

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
            sorting_key: string
                The key to sort the participants by. Can be 'name', 'email', or
                'registrationDate' (the default).
            sorting_direction: string
                Can be either 'ASC' or 'DESC' (the default).
        `Returns:`
            Parsons.Table
                List of participants for the Airmeet event
        """
        participants = []  # List to hold all participants
        page_size = 1000  # Maximum number of results per page

        # Initial API call to get the total user count and first page of data
        response = self.client.get_request(
            url=f"airmeet/{airmeet_id}/participants",
            params={
                "pageNumber": 1,
                "resultSize": page_size,
                "sortingKey": sorting_key,
                "sortingDirection": sorting_direction,
            },
        )
        participants.extend(response["participants"])

        # Calculate total pages needed based on totalUserCount.
        total_count = response["totalUserCount"]
        total_pages = (total_count + page_size - 1) // page_size  # This rounds up the division.

        # Fetch subsequent pages if needed.
        for page in range(2, total_pages + 1):
            response = self.client.get_request(
                url=f"airmeet/{airmeet_id}/participants",
                params={
                    "pageNumber": page,
                    "resultSize": page_size,
                    "sortingKey": sorting_key,
                    "sortingDirection": sorting_direction,
                },
            )
            participants.extend(response["participants"])

        return Table(participants)

    def fetch_airmeet_sessions(self, airmeet_id) -> Table:
        """
        Get the list of sessions for an Airmeet.

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
        `Returns:`
            Parsons.Table
                List of sessions for this Airmeet event
        """
        response = self.client.get_request(url=f"airmeet/{airmeet_id}/info")
        return Table(response["sessions"])

    def fetch_airmeet_info(self, airmeet_id, lists_to_tables=False):
        """
        Get the data for an Airmeet (event), which include the list of
        sessions, session hosts/cohosts, and various other info.

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
            lists_to_tables: bool
                If True, will convert any dictionary values that are lists
                to Tables.
        `Returns:`
            Dict containing the Airmeet data
        """
        response = self.client.get_request(url=f"airmeet/{airmeet_id}/info")
        if lists_to_tables:
            for k in response:
                if isinstance(response[k], list):
                    response[k] = Table(response[k])
        return response

    def fetch_airmeet_custom_registration_fields(self, airmeet_id) -> Table:
        """
        Get the list of custom registration fields for an Airmeet.

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
        `Returns:`
            Parsons.Table
                List of custom registration fields for this Airmeet event
        """
        response = self.client.get_request(url=f"airmeet/{airmeet_id}/custom-fields")
        return Table(response["customFields"])

    def fetch_event_attendance(self, airmeet_id) -> Table:
        """
        Get all attendees for an Airmeet, handling pagination based on the
        returned pageCount.

        Results include attendance only from sessions with a status of
        `FINISHED`. Maximum number of results per page = 50.

        "This is an Asynchronous API. If you get a 202 code in response,
        please try again after 5 minutes."

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
        `Returns:`
            Parsons.Table
                List of attendees for this Airmeet event
        """
        return self._get_all_pages(url=f"airmeet/{airmeet_id}/attendees", page_size=50)

    def fetch_session_attendance(self, session_id) -> Table:
        """
        Get all attendees for a specific Airmeet session, handling pagination
        based on the returned pageCount.

        Results are available only for sessions with a status of `FINISHED`.
        Maximum number of results per page = 50.

        "This is an Asynchronous API. If you get a 202 code in response,
        please try again after 5 minutes."

        `Args:`
            session_id: string
                The id of the session.
        `Returns:`
            Parsons.Table
                List of attendees for this session
        """
        return self._get_all_pages(url=f"session/{session_id}/attendees", page_size=50)

    def fetch_airmeet_booths(self, airmeet_id) -> Table:
        """
        Get the list of booths for a specific Airmeet by ID.

        `CAUTION: This method is untested. Booths are available only in
        certain Airmeet plans.`

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
        `Returns:`
            Parsons.Table
                List of booths for this Airmeet
        """
        response = self.client.get_request(url=f"airmeet/{airmeet_id}/booths")
        return Table(response["booths"] or [])

    def fetch_booth_attendance(self, airmeet_id, booth_id) -> Table:
        """
        Get all attendees for a specific Airmeet booth, handling pagination
        based on the returned pageCount.

        Results are available only for events with a status of `FINISHED`.
        Maximum number of results per page = 50.

        "This is an Asynchronous API. If you get a 202 code in response,
        please try again after 5 minutes."

        `CAUTION: This method is untested. Booths are available only in
        certain Airmeet plans.`

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
            booth_id: string
                The id of the booth.
        `Returns:`
            Parsons.Table
                List of attendees for this booth
        """
        return self._get_all_pages(
            url=f"airmeet/{airmeet_id}/booth/{booth_id}/booth-attendance", page_size=50
        )

    def fetch_poll_responses(self, airmeet_id) -> Table:
        """
        Get a list of the poll responses in an Airmeet, handling pagination
        based on the returned pageCount.

        Maximum number of results per page = 50.

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
        `Returns:`
            Parsons.Table
                List of users. For each user, the value for the "polls"
                key is a list of poll questions and answers for that user.
        """
        return self._get_all_pages(url=f"airmeet/{airmeet_id}/polls", page_size=50)

    def fetch_questions_asked(self, airmeet_id) -> Table:
        """
        Get a list of the questions asked in an Airmeet.

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
        `Returns:`
            Parsons.Table
                List of users. For each user, the value for the "questions"
                key is a list of that user's questions.
        """
        response = self.client.get_request(url=f"airmeet/{airmeet_id}/questions")
        return Table(response["data"])

    def fetch_event_tracks(self, airmeet_id) -> Table:
        """
        Get a list of the tracks in a specific Airmeet by ID.

        `CAUTION: This method is untested. Event tracks are available only in
        certain Airmeet plans.`

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
        `Returns:`
            Parsons.Table
                List of event tracks
        """
        response = self.client.get_request(url=f"airmeet/{airmeet_id}/tracks")
        return Table(response["tracks"])

    def fetch_registration_utms(self, airmeet_id) -> Table:
        """
        Get all the UTM parameters captured during registration, handling
        pagination based on the returned pageCount.

        Maximum number of results per page = ?? (documentation doesn't say,
        but assume 50 like the other asynchronous APIs).

        "This is an Asynchronous API. If you get a 202 code in response,
        please try again after 5 minutes."

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
        `Returns:`
            Parsons.Table
                List of UTM parameters captured during registration
        """
        return self._get_all_pages(url=f"airmeet/{airmeet_id}/utms", page_size=50)

    def download_session_recordings(self, airmeet_id, session_id=None) -> Table:
        """
        Get a list of recordings for a specific Airmeet (and optionally a
        specific session in that Airmeet). The data for each recording
        includes a download link which is valid for 6 hours.

        The API returns "recordingsCount" and "totalCount", which implies
        that the results could be paged like in fetch_airmeet_participants().
        The API docs don't specify if that's the case, but this method will
        need to be updated if it is.

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
            session_id: string
                (optional) If provided, limits results to only the recording
                of the specified session.
        `Returns:`
            Parsons.Table
                List of session recordings
        """
        kwargs = {}
        if session_id:
            kwargs["sessionIds"] = session_id
        response = self.client.get_request(url=f"airmeet/{airmeet_id}/session-recordings", **kwargs)
        return Table(response["recordings"])

    def fetch_event_replay_attendance(self, airmeet_id, session_id=None) -> Table:
        """
        Get all replay attendees for a specific Airmeet (and optionally a
        specific session in that Airmeet), handling pagination based on the
        returned pageCount.

        Results are available only for events with a status of `FINISHED`.
        Maximum number of results per page = 50.

        "This is an Asynchronous API. If you get a 202 code in response,
        please try again after 5 minutes."

        `Args:`
            airmeet_id: string
                The id of the Airmeet.
            session_id: string
                (optional) If provided, limits results to only attendees of
                the specified session.
        `Returns:`
            Parsons.Table
                List of event replay attendees
        """
        attendees = self._get_all_pages(
            url=f"airmeet/{airmeet_id}/event-replay-attendees", page_size=50
        )
        if session_id is not None:
            attendees = attendees.select_rows("{session_id} == '" + session_id + "'")
        return attendees



"""Core methods for running dbt commands."""

import logging
import pathlib
from typing import List, Optional, Type, Union

from dbt.cli.main import dbtRunner, dbtRunnerResult

from parsons.utilities.dbt.logging import dbtLogger
from parsons.utilities.dbt.models import Manifest

logger = logging.getLogger(__name__)


class dbtRunnerParsons:
    def __init__(
        self,
        commands: Union[str, List[str]],
        dbt_project_directory: pathlib.Path,
    ) -> None:
        """Initialize dbtRunner with commands and a working directory.

        `Args:`
            commands: Union[str, List[str]]
                A single dbt command string or a list of dbt command
                strings.
                e.g. ["seed", "build -s models/staging", "test"]
            dbt_project_directory: pathlib.Path
                The path to find the dbt project, as a working
                directory for dbt commands to run
        """
        if isinstance(commands, str):
            commands = [commands]
        self.commands = commands
        self.dbt_project_directory = dbt_project_directory

    def run(self) -> list[Manifest]:
        """Executes dbt commands one by one, returns all results."""
        results = []

        for command in self.commands:
            result = self.execute_dbt_command(command)
            results.append(result)

        return results

    def execute_dbt_command(self, command: str) -> Manifest:
        """Runs dbt command and logs results after process is completed."""
        if command.startswith("dbt "):
            command = command[4:]

        # initialize
        dbt = dbtRunner()

        # create CLI args as a list of strings
        cli_args = command.split(" ")
        cli_args.extend(["--project-dir", str(self.dbt_project_directory)])

        # run the command
        result: dbtRunnerResult = dbt.invoke(cli_args)
        manifest = Manifest(command=command, dbt_manifest=result.result)

        if result.exception:
            raise result.exception

        return manifest


def run_dbt_commands(
    commands: Union[str, List[str]],
    dbt_project_directory: pathlib.Path,
    loggers: Optional[list[Union[dbtLogger, Type[dbtLogger]]]] = None,
) -> list[Manifest]:
    """Executes dbt commands within a directory, optionally logs results.

    Parameters:
    -----------
    commands : Union[str, List[str]]
        A single dbt command as a string or a list of dbt commands to
        be executed.

    dbt_project_directory : pathlib.Path
        The path to the dbt project directory where the commands will
        be executed.

    loggers : Optional[list[Union[dbtLogger, Type[dbtLogger]]]], default=None
        A list of logger instances or logger classes. If classes are
        provided, they will be instantiated.  Each logger should have
        a `send` method that takes the dbt command results as an
        argument.

    Returns:
    --------
    list[Manifest]
        A list of result objects from the executed dbt commands.

    Example:
    --------
    >>> from pathlib import Path
    >>> from parsons.utilities.dbt import (
    ...     run_dbt_commands,
    ...     dbtLoggerSlack,
    ...     dbtLoggerPython
    ... )
    >>> results = run_dbt_commands(
    ...     commands=["dbt run", "dbt test"],
    ...     dbt_project_directory=Path("/path/to/dbt/project"),
    ...     loggers=[dbtLoggerPython, dbtLoggerSlack]
    ... )
    """
    dbt_runner = dbtRunnerParsons(commands, dbt_project_directory)

    dbt_command_results = dbt_runner.run()

    if loggers:
        for logger in loggers:
            if not isinstance(logger, dbtLogger):
                # Instantiate logger if not already instantiated
                logger = logger()
            logger.send(dbt_command_results)

    return dbt_command_results



import logging

import psycopg2
import sshtunnel


def query_through_ssh(
    ssh_host,
    ssh_port,
    ssh_username,
    ssh_password,
    db_host,
    db_port,
    db_name,
    db_username,
    db_password,
    query,
):
    """
    `Args:`
        ssh_host:
            The host for the SSH connection
        ssh_port:
            The port for the SSH connection
        ssh_username:
            The username for the SSH connection
        ssh_password:
            The password for the SSH connection
        db_host:
            The host for the db connection
        db_port:
            The port for the db connection
        db_name:
            The name of the db database
        db_username:
            The username for the db database
        db_password:
            The password for the db database
        query:
            The SQL query to execute

    `Returns:`
        A list of records resulting from the query or None if something went wrong
    """
    output = None
    server = None
    con = None
    try:
        server = sshtunnel.SSHTunnelForwarder(
            (ssh_host, int(ssh_port)),
            ssh_username=ssh_username,
            ssh_password=ssh_password,
            remote_bind_address=(db_host, int(db_port)),
        )
        server.start()
        logging.info("SSH tunnel established successfully.")

        con = psycopg2.connect(
            host="localhost",
            port=server.local_bind_port,
            database=db_name,
            user=db_username,
            password=db_password,
        )
        logging.info("Database connection established successfully.")

        cursor = con.cursor()
        cursor.execute(query)
        records = cursor.fetchall()
        output = records
        logging.info(f"Query executed successfully: {records}")
    except Exception as e:
        logging.error(f"Error during query execution: {e}")
        raise e
    finally:
        if con:
            con.close()
            logging.info("Database connection closed.")
        if server:
            server.stop()
            logging.info("SSH tunnel closed.")
    return output



import logging

from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

COMMUNITY_API_ENDPOINT = "https://dl.community.com/download/v1/files/"


class Community(object):
    """
    Instantiate class.

       `Args:`
            community_client_id: str
                The Community provided Client ID. Not required if ``COMMUNITY_CLIENT_ID`` env
                variable set.
            community_access_token: str
                The Community provided access token. Not required if ``COMMUNITY_ACCESS_TOKEN`` env
                variable set.
            community_uri: str
                The URI to access the  API. Not required, default is
                `<https://dl.community.com/download/v1/files/>`_. You can set an ``COMMUNITY_URL`` env
                variable or use this URI parameter if a different endpoint is necessary.

    `API Documentation <https://developer.community.com/reference/data-export-api-downloading-data>`_
    """

    def __init__(self, community_client_id=None, community_access_token=None, community_url=None):
        self.community_client_id = check_env.check("community_client_id", community_client_id)
        self.community_access_token = check_env.check(
            "community_access_token", community_access_token
        )
        self.uri = (
            check_env.check("COMMUNITY_URL", community_url, optional=True)
            or f"{COMMUNITY_API_ENDPOINT}/{community_client_id}/"
        )
        self.headers = {
            "Authorization": f"Bearer {self.community_access_token}",
        }
        self.client = APIConnector(
            self.uri,
            headers=self.headers,
        )

    def get_request(self, filename):
        """
        GET request to Community.com API to get the CSV data.

        `Args:`
            filename: str
                Data filename you are requesting.
                Options:
                    'campaigns': Campaign Performance data
                    'outbound_message_type_usage`: Message Segment Usage data
                    'campaign_links': Campaign Link Performance data
                    'members': Member Details data
                    'member_state_changes': Member Subscription Status data
                    'custom_member_data': Custom Member Data
                    'communities': Communities data
                    'member_communities': Member Communities data

        `Returns:`
            Response of GET request; a successful response returns the CSV formatted data
        """

        logger.info(f"Requesting {filename}")
        url = (
            f"{filename}.csv.gz"
            if filename != "outbound_message_type_usage"
            else f"{filename}.csv.gz/segment-based-subscription"
        )
        response = self.client.get_request(url=url, return_format="content")
        return response

    def get_data_export(self, filename):
        """
        Get specified data from Community.com API as Parsons table.

        `Args:`
            filename: str
                Data filename you are requesting.
                Options:
                    'campaigns': Campaign Performance data
                    'outbound_message_type_usage`: Message Segment Usage data
                    'campaign_links': Campaign Link Performance data
                    'members': Member Details data
                    'member_state_changes': Member Subscription Status data
                    'custom_member_data': Custom Member Data
                    'communities': Communities data
                    'member_communities': Member Communities data

        `Returns:`
            Contents of the generated contribution CSV as a Parsons table.
        """

        get_request_response = self.get_request(filename=filename)
        response_string = get_request_response.decode("utf-8")
        table = Table.from_csv_string(response_string)
        return table
